# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

AGENT_TYPE_OVSVAPP = "OVSvApp L2 Agent"
OVSVAPP = 'ovsvapp'
DEVICE = 'device'

# Network type constants.
NETWORK_VLAN = 'VLAN'
NETWORK_VXLAN = 'VXLAN'

# Port status constants.
PORT_STATUS_UP = 'UP'
PORT_STATUS_DOWN = 'DOWN'

# VM Event type constants.
VM_CREATED = 'VM_CREATED'
VM_UPDATED = 'VM_UPDATED'
VM_DELETED = 'VM_DELETED'

# Driver state constants.
# Driver cannot connect or monitor hypervisor. Driver cannot process
# ResourceEntity APIs.
DRIVER_IDLE = 'IDLE'
# Driver is ready for monitoring hypervisor.
DRIVER_READY = 'READY'
# Driver is monitoring hypervisor.
DRIVER_RUNNING = 'RUNNING'
# Driver is stopped.
DRIVER_STOPPED = 'STOPPED'

# VLAN ID constants.
DEAD_VLAN = 4095

# Agent State constants.
AGENT_INITIALIZING = "INITIALIZING"
AGENT_INITIALIZED = "INITIALIZED"
AGENT_RUNNING = "RUNNING"
AGENT_STOPPING = "STOPPING"
AGENT_STOPPED = "STOPPED"

SEC_TO_INT_PATCH = "patch-integration"
INT_TO_SEC_PATCH = "patch-security"

# OVS Firewall related constants.
SG_DROPALL_PRI = 0
SG_DEFAULT_PRI = 1
SG_LOW_PRI = 5
SG_RULES_PRI = 10
SG_TP_PRI = 20
SG_TCP_FLAG_PRI = 25
SG_DROP_HIGH_PRI = 50

SG_DEFAULT_TABLE_ID = 0
SG_IP_TABLE_ID = 2
SG_TCP_TABLE_ID = 2
SG_UDP_TABLE_ID = 2
SG_ICMP_TABLE_ID = 2
SG_LEARN_TABLE_ID = 5

ICMP_ECHO_REQ = 8
ICMP_ECHO_REP = 0
ICMP_TIME_EXCEEDED = 11
ICMP_TS_REQ = 13
ICMP_TS_REP = 14
ICMP_INFO_REQ = 15
ICMP_INFO_REP = 16
ICMP_AM_REQ = 17
ICMP_AM_REP = 18
ICMP_DEST_UNREACH = 3



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log

from networking_vsphere.common import constants
from networking_vsphere.common import utils
from networking_vsphere.drivers import base_manager as manager
from networking_vsphere.drivers import driver

LOG = log.getLogger(__name__)


class Agent(driver.NetworkDriverCallback):

    """Agent class.

    Base class for agents - initializing, starting and
    stopping driver manager.
    """

    def __init__(self):
        self.net_mgr = None
        self.state = constants.AGENT_INITIALIZING
        self.node_up = False

    def _stop_managers(self):
        LOG.debug("Stopping managers.")
        self.state = constants.AGENT_STOPPING
        if self.net_mgr:
            self.net_mgr.stop()
        self.state = constants.AGENT_STOPPED

    def _initialize_managers(self):
        self.state = constants.AGENT_INITIALIZING
        LOG.info(_("Loading network driver manager: %s."),
                 cfg.CONF.OVSVAPP.network_manager)
        self.net_mgr = utils.load_object(cfg.CONF.OVSVAPP.network_manager,
                                         manager.DriverManager,
                                         self)
        self.net_mgr.initialize_driver()
        self.state = constants.AGENT_INITIALIZED

    def _start_managers(self):
        if self.state == constants.AGENT_INITIALIZED and self.node_up:
            LOG.debug("Starting managers.")
            if self.net_mgr:
                self.net_mgr.start()
            self.state = constants.AGENT_RUNNING

    def set_node_state(self, is_up):
        if is_up != self.node_up:
            self.node_up = is_up
            if is_up:
                LOG.info(_("Making node up."))
                self._initialize_managers()
                self._start_managers()
            else:
                self.state = constants.AGENT_INITIALIZING
                self._stop_managers()
        else:
            LOG.info(_("Ignoring node update as agent "
                       "is already %s."),
                     "ACTIVE" if self.node_up else "DOWN")



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_log import log
from oslo_vmware import api
from oslo_vmware import vim

LOG = log.getLogger(__name__)


class ConnectionHandler:
    session = None
    host_ip = None
    host_username = None
    host_password = None
    api_retry_count = 0
    wsdl_url = None
    scheme = None
    stopped = False
    create_session = True
    https_port = 443
    ca_cert = False

    @classmethod
    def set_vc_details(cls, host_ip, host_username, host_password,
                       api_retry_count, wsdl_url, https_port=443,
                       scheme="https"):
        cls.session = None
        cls.host_ip = host_ip
        cls.host_username = host_username
        cls.host_password = host_password
        cls.api_retry_count = api_retry_count
        cls.wsdl_url = wsdl_url
        cls.scheme = scheme
        cls.https_port = https_port
        cls.stopped = False
        cls.create_session = True
        cls.ca_cert = False

    @classmethod
    def stop(cls):
        cls.stopped = True
        if cls.session:
            cls.session.logout()
        cls.session = None

    @classmethod
    def start(cls):
        cls.stopped = False

    @classmethod
    def create_connection(cls):
        cls.session = VMWareAPISession(cls.host_ip,
                                       cls.host_username,
                                       cls.host_password,
                                       cls.api_retry_count,
                                       cls.wsdl_url,
                                       scheme=cls.scheme,
                                       https_port=cls.https_port,
                                       create_session=cls.create_session,
                                       ca_cert=cls.ca_cert)
        return cls.session

    @classmethod
    def get_connection(cls, create=False):
        if not cls.session and create:
            return cls.create_connection()
        else:
            return cls.session

    @classmethod
    def try_connection(cls):
        while not cls.stopped:
            try:
                return cls.get_connection(create=True)
            except Exception as e:
                LOG.error(_("Connection to vCenter %(host_ip)s failed -"
                            "%(exception)s"), {"host_ip": cls.host_ip,
                                               "exception": e})
                LOG.error(_("Will retry after 60 secs"))
                time.sleep(60)
                LOG.error(_("Retrying VMWare Connection after 60 secs"))
                continue


class VMWareAPISession(api.VMwareAPISession):

    """Sets up a session with the ESX host and handles all the calls."""

    def __init__(self, host_ip, host_username, host_password,
                 api_retry_count, wsdl_url, scheme="https", https_port=443,
                 create_session=True, ca_cert=None):
        super(VMWareAPISession, self).__init__(
            host=host_ip,
            port=https_port,
            server_username=host_username,
            server_password=host_password,
            api_retry_count=api_retry_count,
            scheme=scheme,
            task_poll_interval=1,
            wsdl_loc=wsdl_url,
            create_session=create_session,
            cacert=ca_cert)

    def __del__(self):
        """Logs-out the session."""
        # Logout to avoid un-necessary increase in session count at the
        # ESX host
        try:
            self.logout()
        except Exception:
            # It is just cautionary on our part to do a logout in del just
            # to ensure that the session is not left active.
            pass

    def _is_vim_object(self, module):
        """Check if the module is a VIM Object instance."""
        return isinstance(module, vim.Vim)

    def _call_method(self, module, method, *args, **kwargs):
        """Calls a method within the module specified with args provided."""
        if not self._is_vim_object(module):
            return self.invoke_api(module, method, self.vim, *args, **kwargs)
        else:
            return self.invoke_api(module, method, *args, **kwargs)

    def _get_vim(self):
        """Gets the VIM object reference."""
        return self.vim



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_vsphere.tests import base
from networking_vsphere.utils import error_util


class TestVimException(base.TestCase):

    def test_str(self):
        exc = Exception("Test Exception")
        exception_summary = "Test Summary "
        vim_excep = error_util.VimException(exception_summary, exc)
        self.assertEqual(str(exception_summary) + str(exc), str(vim_excep))


class TestVimFaultException(base.TestCase):

    def test_str(self):
        exc = Exception("Test Exception")
        fault_list = ["NotAuthenticated"]
        vim_fault = error_util.VimFaultException(fault_list, exc)
        self.assertEqual(str(exc), str(vim_fault))


# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock

from networking_vsphere.common import constants
from networking_vsphere.tests import base
from networking_vsphere.tests.unit.utils import fake_vmware_api as fake_api
from networking_vsphere.tests.unit.utils import stubs
from networking_vsphere.utils import error_util
from networking_vsphere.utils import network_util
from networking_vsphere.utils import resource_util
from networking_vsphere.utils import vim_util


class TestVmwareNetworkUtil(base.TestCase):

    def setUp(self):
        super(TestVmwareNetworkUtil, self).setUp()
        self.fake_visdk = self.useFixture(stubs.FakeVmware())
        self.session = self.fake_visdk.session
        self.useFixture(stubs.CacheFixture())

    def test_get_dvs_mor_by_uuid(self):
        self.assertTrue(
            network_util.get_dvs_mor_by_uuid(self.session, "fake_dvs"))

    def test_get_dvs_mor_by_non_uuid(self):
        self.assertFalse(
            network_util.get_dvs_mor_by_uuid(self.session, "invalid_dvs"))

    def test_get_dvs_mor_by_name(self):
        self.assertTrue(
            network_util.get_dvs_mor_by_name(self.session, "test_dvs"))

    def test_get_dvs_mor_by_name_for_invalid_dvs(self):
        self.assertFalse(
            network_util.get_dvs_mor_by_name(self.session, "invalid_dvs"))

    def test_get_portgroup_mor_by_name(self):
        dvs_name = "test_dvs"
        port_group_name = fake_api.Constants.PORTGROUP_NAME
        dvs = fake_api.DataObject()
        dvs_config = fake_api.DataObject()
        port_group_mors = []
        pg1 = fake_api.create_network()
        pg1.set("summary.name", "pg1")
        port_group_mors.append(pg1)
        pg2 = fake_api.create_network()
        pg2.set("summary.name", port_group_name)
        port_group_mors.append(pg2)
        dvs_config.ManagedObjectReference = port_group_mors
        with contextlib.nested(
            mock.patch.object(vim_util,
                              'get_properties_for_a_collection_of_objects',
                              return_value=port_group_mors),
            mock.patch.object(network_util, "get_dvs_mor_by_name",
                              return_value=dvs),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=dvs_config)):
                port_group = network_util.get_portgroup_mor_by_name(
                    self.session, dvs_name, port_group_name)
                self.assertEqual(port_group.value, pg2.value)

    def test_get_portgroup_mor_by_name_no_dvs(self):
        dvs_name = "non_existent_dvs"
        port_group_name = fake_api.Constants.PORTGROUP_NAME
        with mock.patch.object(network_util, "get_dvs_mor_by_name",
                               return_value=None):
            port_group = network_util.get_portgroup_mor_by_name(
                self.session, dvs_name, port_group_name)
            self.assertIsNone(port_group)

    def test_get_portgroup_mor_by_name_not_found(self):
        dvs_name = "test_dvs"
        port_group_name = fake_api.Constants.PORTGROUP_NAME
        dvs = fake_api.DataObject()
        dvs_config = fake_api.DataObject()
        port_group_mors = []
        pg1 = fake_api.create_network()
        pg1.set("summary.name", "pg1")
        port_group_mors.append(pg1)
        dvs_config.ManagedObjectReference = port_group_mors
        with contextlib.nested(
            mock.patch.object(vim_util,
                              'get_properties_for_a_collection_of_objects',
                              return_value=port_group_mors),
            mock.patch.object(network_util, "get_dvs_mor_by_name",
                              return_value=dvs),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=dvs_config)):
                port_group = network_util.get_portgroup_mor_by_name(
                    self.session, dvs_name, port_group_name)
                self.assertIsNone(port_group)

    def test_get_all_portgroup_mors_for_switch(self):
        port_group_mors = network_util.get_all_portgroup_mors_for_switch(
            self.session, "test_dvs")
        self.assertTrue(port_group_mors)
        self.assertTrue(isinstance(port_group_mors, list))

    def test_get_all_portgroup_mors_for_invalid_switch(self):
        dvs_name = "test_invalid_dvs"
        with mock.patch.object(network_util, "get_dvs_mor_by_name",
                               return_value=None):
            self.assertFalse(network_util.get_all_portgroup_mors_for_switch
                             (self.session, dvs_name))

    def test_get_unused_portgroup_names(self):
        dvp = 'DistributedVirtualPortgroup'
        fake_api._db_content[dvp].values()[0].propSet[1].val = None
        self.assertTrue(network_util.get_unused_portgroup_names(self.session,
                                                                "test_dvs"))

    def test_get_used_portgroup_names(self):
        self.assertFalse(network_util.get_unused_portgroup_names(self.session,
                                                                 "test_dvs"))

    def test_get_portgroup_details(self):
        res = network_util.get_portgroup_details(self.session,
                                                 "test_dvs",
                                                 fake_api.Constants.
                                                 PORTGROUP_NAME)
        self.assertEqual(100, res)

    def test_get_portgroup_details_not_found(self):
        with mock.patch.object(network_util, "get_portgroup_mor_by_name",
                               return_value=None):
            self.assertEqual(network_util.get_portgroup_details(
                self.session,
                "test_invalid_dvs",
                fake_api.Constants.PORTGROUP_NAME), constants.DEAD_VLAN)

    def test_create_port_group_existing(self):
        dvs_name = "test_dvs"
        pg_name = fake_api.Constants.PORTGROUP_NAME
        vlanid = "100"
        pg = fake_api.DataObject()
        defaultPortConfig = fake_api.DataObject()
        vlan = fake_api.DataObject()
        vlan.vlanId = vlanid
        defaultPortConfig.vlan = vlan
        port_group_config = fake_api.DataObject()
        port_group_config.defaultPortConfig = defaultPortConfig
        with contextlib.nested(
            mock.patch.object(network_util, "get_portgroup_mor_by_name",
                              return_value=pg),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=port_group_config)
        ) as (mor, get_prop):
                network_util.create_port_group(self.session, dvs_name, pg_name,
                                               vlanid)
                self.assertTrue(get_prop.called)

    def test_create_port_group_with_invalid_vlanid(self):
        dvs_name = "test_dvs"
        pg_name = fake_api.Constants.PORTGROUP_NAME
        vlanid = "100"
        pg = fake_api.DataObject()
        defaultPortConfig = fake_api.DataObject()
        vlan = fake_api.DataObject()
        vlan.vlanId = "200"
        defaultPortConfig.vlan = vlan
        port_group_config = fake_api.DataObject()
        port_group_config.defaultPortConfig = defaultPortConfig
        with contextlib.nested(
            mock.patch.object(network_util, "get_portgroup_mor_by_name",
                              return_value=pg),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=port_group_config)):
                raised = self.assertRaises(error_util.RunTimeError,
                                           network_util.create_port_group,
                                           self.session,
                                           dvs_name, pg_name, vlanid)
                self.assertTrue(raised)

    def test_create_port_group_err_status(self):
        dvs_name = "test_dvs"
        pg_name = fake_api.Constants.PORTGROUP_NAME
        vlanid = "5001"
        task_info = fake_api.DataObject()
        task_info.name = "AddDVPortgroup_Task"
        task_info.key = "task-1234"
        task_info.state = "error"
        task_info.error = fake_api.DataObject()
        with contextlib.nested(
            mock.patch.object(network_util, "get_portgroup_mor_by_name",
                              return_value=None),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=task_info)):
                raised = self.assertRaises(error_util.RunTimeError,
                                           network_util.create_port_group,
                                           self.session, dvs_name,
                                           pg_name, vlanid)
                self.assertTrue(raised)

    def test_wait_until_dvs_portgroup_available(self):
        vm_ref = fake_api._db_content["VirtualMachine"].values()[0].obj
        hs_key = fake_api._db_content["HostSystem"].keys()[0]
        dvs_key = fake_api._db_content[
            "DistributedVirtualPortgroup"].keys()[0]
        dvs_obj = fake_api._db_content[
            "DistributedVirtualPortgroup"][dvs_key].obj
        network_obj = fake_api.DataObject()
        network_obj.name = fake_api.Constants.PORTGROUP_NAME
        network_obj.ManagedObjectReference = [dvs_obj]
        fake_api._db_content["HostSystem"][
            hs_key].propSet[2].val = network_obj
        self.assertTrue(network_util.wait_until_dvs_portgroup_available(
            self.session,
            vm_ref,
            fake_api.Constants.PORTGROUP_NAME,
            3))

    def test_wait_until_dvs_portgroup_unavailable(self):
        vm_ref = fake_api._db_content["VirtualMachine"].values()[0].obj
        hs_key = fake_api._db_content["HostSystem"].keys()[0]
        dvs_key = fake_api._db_content[
            "DistributedVirtualPortgroup"].keys()[0]
        dvs_obj = fake_api._db_content["DistributedVirtualPortgroup"][
            dvs_key].obj
        network_obj = fake_api.DataObject()
        network_obj.name = fake_api.Constants.PORTGROUP_NAME
        network_obj.ManagedObjectReference = [dvs_obj]
        fake_api._db_content["HostSystem"][
            hs_key].propSet[2].val = network_obj
        self.assertFalse(network_util.wait_until_dvs_portgroup_available(
            self.session,
            vm_ref,
            "invalid_pg",
            3))

    def test_delete_port_group(self):
        self.assertTrue(network_util.get_portgroup_mor_by_name(
            self.session, "test_dvs",
            fake_api.Constants.PORTGROUP_NAME))
        network_util.delete_port_group(self.session, "test_dvs",
                                       fake_api.Constants.PORTGROUP_NAME)
        self.assertFalse(network_util.get_portgroup_mor_by_name(
            self.session, "test_dvs", fake_api.Constants.PORTGROUP_NAME))

    def test_delete_port_group_invalid_dvs(self):
        with mock.patch.object(self.session, "wait_for_task") as task_wait:
            network_util.delete_port_group(self.session,
                                           "test_invalid_dvs",
                                           fake_api.Constants.PORTGROUP_NAME)
            self.assertFalse(task_wait.called)

    def test_delete_port_group_exc(self):
        dvs_name = "test_dvs"
        pg_name = fake_api.Constants.PORTGROUP_NAME
        task_info = fake_api.DataObject()
        task_info.name = "Destroy_Task"
        task_info.key = "task-777"
        task_info.state = "error"
        task_info.error = fake_api.DataObject()
        with contextlib.nested(
            mock.patch.object(network_util, "get_portgroup_mor_by_name",
                              return_value=True),
            mock.patch.object(vim_util, "get_dynamic_property",
                              return_value=task_info)):
                raised = self.assertRaises(error_util.RunTimeError,
                                           network_util.delete_port_group,
                                           self.session, dvs_name,
                                           pg_name)
                self.assertTrue(raised)

    def test_enable_disable_port_of_vm_existing(self):
        pg = fake_api._db_content[
            "DistributedVirtualPortgroup"].values()[0]
        backing = fake_api.DataObject()
        backing.port = fake_api.DataObject()
        backing.port.portgroupKey = pg.value
        backing.port.portKey = pg.portKeys[0]
        backing.port.switchUuid = fake_api._db_content[
            "DistributedVirtualPortgroup"].keys()[0]
        vm_key = fake_api._db_content["VirtualMachine"].keys()[0]
        fake_api._db_content["VirtualMachine"][vm_key].propSet[
            6].val.VirtualDevice[0].backing = backing
        self.assertTrue(network_util.enable_disable_port_of_vm(
            self.session,
            fake_api._db_content["VirtualMachine"].values()[0],
            "00:99:88:77:66:ab",
            True))

    def test_enable_disable_port_of_vm_non_existing(self):
        pg = fake_api._db_content[
            "DistributedVirtualPortgroup"].values()[0]
        backing = fake_api.DataObject()
        backing.port = fake_api.DataObject()
        backing.port.portgroupKey = pg.value
        backing.port.portKey = pg.portKeys[0]
        backing.port.switchUuid = fake_api._db_content[
            "DistributedVirtualPortgroup"].keys()[0]
        vm_key = fake_api._db_content["VirtualMachine"].keys()[0]
        fake_api._db_content["VirtualMachine"][vm_key].propSet[
            6].val.VirtualDevice[0].backing = backing
        self.assertFalse(network_util.enable_disable_port_of_vm(
            self.session,
            fake_api._db_content["VirtualMachine"].values()[0],
            "11:99:88:77:66:ab",
            True))

    def test_is_valid_dvswitch(self):
        cluster_mor = resource_util.get_cluster_mor_for_vm(
            self.session, fake_api.Constants.VM_UUID)
        self.assertTrue(network_util.is_valid_dvswitch(self.session,
                                                       cluster_mor,
                                                       "test_dvs"))

    def test_is_valid_dvswitch_no_dvs(self):
        cluster_mor = resource_util.get_cluster_mor_for_vm(
            self.session, fake_api.Constants.VM_UUID)
        self.assertFalse(network_util.is_valid_dvswitch(self.session,
                                                        cluster_mor,
                                                        "invalid_dvs"))

    def test_is_valid_dvswitch_no_host(self):
        cluster_mor = resource_util.get_cluster_mor_for_vm(
            self.session, fake_api.Constants.VM_UUID)
        with mock.patch.object(resource_util, "get_host_mors_for_cluster",
                               return_value=None):
            self.assertFalse(network_util.is_valid_dvswitch(self.session,
                                                            cluster_mor,
                                                            "test_dvs"))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_log import log
import oslo_messaging
from sqlalchemy.orm import exc as sa_exc

from networking_vsphere.common import constants as ovsvapp_const
from networking_vsphere.db import ovsvapp_db

from neutron.common import exceptions as exc
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.db import models_v2
from neutron.extensions import portbindings
from neutron import manager
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2 import driver_context

LOG = log.getLogger(__name__)


class OVSvAppServerRpcCallback(object):

    """Plugin side of the OVSvApp rpc.

    This class contains extra rpc callbacks to be served for use by the
    OVSvApp Agent.
    """
    target = oslo_messaging.Target(version='1.0')

    def __init__(self, notifier=None):
        super(OVSvAppServerRpcCallback, self).__init__()
        self.notifier = notifier

    @property
    def plugin(self):
        return manager.NeutronManager.get_plugin()

    def _get_devices_info(self, devices):
        return dict(
            (port['id'], port)
            for port in self.plugin.get_ports_from_devices(devices)
            if port and not port['device_owner'].startswith('network:')
        )

    def get_ports_for_device(self, rpc_context, **kwargs):
        """RPC for getting port info.

        This method provides information about the network and port for
        a given device_id.
        """
        agent_id = kwargs.get('agent_id')
        host = kwargs.get('host')
        device = kwargs.get('device')
        device_id = device['id']
        vcenter_id = device['vcenter']
        cluster_id = device['cluster_id']
        LOG.debug("Device %(device_id)s details requested by agent "
                  "%(agent_id)s running on host %(host)s.",
                  {'device_id': device_id, 'agent_id': agent_id, 'host': host})
        if not device_id:
            return False
        try_count = 3
        try:
            while try_count > 0:
                ports = self.plugin.get_ports(rpc_context,
                                              filters={'device_id':
                                                       [device_id]})
                device_ports = []
                sg_port_ids = set()
                for port in ports:
                    network = self.plugin.get_network(rpc_context,
                                                      port['network_id'])
                    port.update(
                        {'network_type': network['provider:network_type'],
                         'segmentation_id':
                         network['provider:segmentation_id'],
                         'physical_network':
                         network['provider:physical_network']})

                    if port['network_type'] == 'vxlan':
                        port_info = {'port_id': port['id'],
                                     'vcenter_id': vcenter_id,
                                     'cluster_id': cluster_id,
                                     'network_id': port['network_id']}
                        lvid = ovsvapp_db.get_local_vlan(port_info)
                        if lvid:
                            port['lvid'] = lvid
                        else:
                            # Local VLANs are exhausted ! No point processing
                            # further.
                            LOG.error(_("No VLAN available in the cluster "
                                        "%(cluster)s for assignment to device "
                                        "%(device)s in vCenter %(vcenter)s."),
                                      {'device': device_id,
                                       'cluster': cluster_id,
                                       'vcenter': vcenter_id})
                            return False
                    else:
                        port['lvid'] = port['segmentation_id']
                    # Bind the port here. If binding succeeds, then
                    # add this port to process for security groups, otheriwse
                    # ignore it.
                    updated_port = self.update_port_binding(rpc_context,
                                                            agent_id=agent_id,
                                                            port_id=port['id'],
                                                            host=host)
                    if not updated_port:
                        LOG.error(_("Port binding failed for "
                                    "port %s."), port['id]'])
                        # process the next port for the device
                        continue
                    if 'security_groups' in port:
                        sg_port_ids.add(port['id'])
                    device_ports.append(port)
                if not device_ports:
                    try_count -= 1
                    LOG.debug("Port details could not be retrieved for "
                              "device %s ..retrying.", device_id)
                    time.sleep(3)
                else:
                    LOG.debug("Device details returned by server: "
                              "%s.", device_ports)
                    # Get the SG rules for the security enabled ports.
                    sg_payload = {}
                    if sg_port_ids:
                        ports = self._get_devices_info(sg_port_ids)
                        sg_rules = self.plugin.security_group_rules_for_ports(
                            rpc_context, ports)
                        sg_payload[device_id] = sg_rules
                    self.notifier.device_create(rpc_context, device,
                                                device_ports, sg_payload)
                    return True
        except Exception:
            LOG.exception(_("Failed to retrieve port details for "
                            "device: %s."), device_id)
        LOG.debug("Failed to retrieve ports for device: %s.", device_id)
        return False

    def update_port_binding(self, rpc_context, **kwargs):
        agent_id = kwargs.get('agent_id')
        port_id = kwargs.get('port_id')
        host = kwargs.get('host')
        LOG.debug("Port %(port_id)s update_port_binding() invoked by agent "
                  "%(agent_id)s for host %(host)s.",
                  {'port_id': port_id, 'agent_id': agent_id, 'host': host})
        port = {'port': {portbindings.HOST_ID: host}}
        updated_port = self.plugin.update_port(rpc_context, port_id, port)
        return updated_port

    def update_ports_binding(self, rpc_context, **kwargs):
        agent_id = kwargs.get('agent_id')
        ports = kwargs.get('ports')
        host = kwargs.get('host')
        updated_ports = set()
        for port_id in ports:
            # Update not required if the host hasn't changed.
            if self.plugin.port_bound_to_host(rpc_context, port_id, host):
                updated_ports.add(port_id)
                continue
            LOG.debug("Port %(port_id)s update_port invoked by agent "
                      "%(agent_id)s for host %(host)s.",
                      {'port_id': port_id, 'agent_id': agent_id,
                       'host': host})
            port = {'port': {portbindings.HOST_ID: host}}
            try:
                updated_port = self.plugin.update_port(rpc_context, port_id,
                                                       port)
                updated_ports.add(updated_port['id'])
            except Exception:
                LOG.exception(_("Failed to update binding for port %s "),
                              port_id)
        return updated_ports

    def _get_port_db(self, session, port_id, agent_id):
        try:
            port_db = (session.query(models_v2.Port).
                       enable_eagerloads(False).
                       filter(models_v2.Port.id.startswith(port_id)).
                       one())
            return port_db
        except sa_exc.NoResultFound:
            LOG.warning(_("Port %(port_id)s requested by agent "
                          "%(agent_id)s not found in database."),
                        {'port_id': port_id, 'agent_id': agent_id})
            return None
        except exc.MultipleResultsFound:
            LOG.error(_("Multiple ports have port_id starting with %s."),
                      port_id)
            return None

    def get_ports_details_list(self, rpc_context, **kwargs):
        """Agent requests device details."""
        agent_id = kwargs.get('agent_id')
        port_ids = kwargs.get('port_ids')
        vcenter_id = kwargs['vcenter_id']
        cluster_id = kwargs['cluster_id']
        LOG.debug("Port details requested by agent "
                  "%(agent_id)s for ports %(ports)s.",
                  {'ports': port_ids, 'agent_id': agent_id})
        out_ports = []
        for port_id in port_ids:
            port_db = self._get_port_db(rpc_context.session, port_id, agent_id)
            if not port_db:
                continue
            port = self.plugin._make_port_dict(port_db)
            network = self.plugin.get_network(rpc_context, port['network_id'])
            port_context = driver_context.PortContext(self.plugin,
                                                      rpc_context,
                                                      port,
                                                      network,
                                                      port_db.port_binding)
            segment = port_context.top_bound_segment
            # Reference: ML2  Driver API changes for hierarchical port binding.
            bound_port = port_context.current

            if not segment:
                LOG.warning(_("Port %(port_id)s requested by agent "
                              "%(agent_id)s on network %(network_id)s not "
                              "bound, vif_type: %(vif_type)s."),
                            {'port_id': port['id'],
                             'agent_id': agent_id,
                             'network_id': port['network_id'],
                             'vif_type': port[portbindings.VIF_TYPE]})
                continue
            bound_port['lvid'] = None
            if segment[api.NETWORK_TYPE] == 'vxlan':
                port_info = {'port_id': bound_port['id'],
                             'vcenter_id': vcenter_id,
                             'cluster_id': cluster_id,
                             'network_id': bound_port['network_id']}
                lvid = ovsvapp_db.get_local_vlan(port_info, False)
                if lvid:
                    bound_port['lvid'] = lvid
                else:
                    # Local VLANs are exhausted !! No point processing
                    # further.
                    LOG.error(_("Local VLAN not available in the cluster"
                                " %(cluster)s for port"
                                " %(port_id)s in vcenter %(vcenter)s."),
                              {'port_id': bound_port['id'],
                               'cluster': cluster_id,
                               'vcenter': vcenter_id})
                    continue
                    # Skip sending back this port as there is no lvid.
            else:
                bound_port['lvid'] = segment[api.SEGMENTATION_ID]

            entry = {'network_id': bound_port['network_id'],
                     'port_id': bound_port['id'],
                     'lvid': bound_port['lvid'],
                     'mac_address': bound_port['mac_address'],
                     'admin_state_up': bound_port['admin_state_up'],
                     'network_type': segment[api.NETWORK_TYPE],
                     'segmentation_id': segment[api.SEGMENTATION_ID],
                     'physical_network': segment[api.PHYSICAL_NETWORK],
                     'fixed_ips': bound_port['fixed_ips'],
                     'device_id': bound_port['device_id'],
                     'device_owner': bound_port['device_owner']}
            LOG.debug("Adding port detail: %s.", entry)
            out_ports.append(entry)
        return out_ports

    def update_lvid_assignment(self, rpc_context, **kwargs):
        net_info = kwargs.get('net_info')
        if net_info:
            try:
                ovsvapp_db.release_local_vlan(net_info)
            except Exception:
                LOG.exception(_("Failed to release the local vlan"))
        return


class OVSvAppAgentNotifyAPI(object):

    """Agent side of the OVSvApp rpc API."""

    def __init__(self, topic=topics.AGENT):
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)
        self.topic = topic

    def _get_device_topic(self, action):
        return topics.get_topic_name(self.topic,
                                     ovsvapp_const.DEVICE,
                                     action)

    def device_create(self, context, device, ports, sg_rules):
        cctxt = self.client.prepare(
            topic=self._get_device_topic(topics.CREATE), fanout=True)
        cctxt.cast(context, 'device_create', device=device, ports=ports,
                   sg_rules=sg_rules)

    def device_delete(self, context, network_info, host):
        cctxt = self.client.prepare(
            topic=self._get_device_topic(topics.DELETE))
        return cctxt.call(context, 'device_delete',
                          network_info=network_info, host=host)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class DriverManager(object):

    '''Base class for all driver managers.'''

    def __init__(self):
        self.driver = None

    def get_driver(self):
        return self.driver

    def initialize_driver(self):
        '''Initialize the driver managed by this manager.'''
        raise NotImplementedError()

    def start(self):
        '''Start driver action when node is up.'''
        pass

    def pause(self):
        '''Handle pause action.'''
        pass

    def stop(self):
        '''Handle stop action.'''
        pass



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_vsphere.drivers import driver
from networking_vsphere.tests import base
from networking_vsphere.tests.unit.drivers import fake_driver


class TestNetworkDriver(base.TestCase):

    def setUp(self):
        super(TestNetworkDriver, self).setUp()
        self.driver = driver.NetworkDriver()

    def test_set_callback(self):
        mock_driver = fake_driver.MockNetworkDriver()
        callback = fake_driver.MockCallback()
        mock_driver.set_callback(callback)
        self.assertEqual(mock_driver.callback_impl, callback)

    def test_monitor_events(self):
        self.assertRaises(NotImplementedError,
                          self.driver.monitor_events)

    def test_pause(self):
        self.assertIsNone(self.driver.pause())

    def test_stop(self):
        self.assertIsNone(self.driver.stop())

    def test_is_connected(self):
        self.assertRaises(NotImplementedError,
                          self.driver.is_connected)

    def test_create_network(self):
        self.assertRaises(NotImplementedError,
                          self.driver.create_network, None, None)

    def test_delete_network(self):
        self.assertRaises(NotImplementedError,
                          self.driver.delete_network, None, None)

    def test_prepare_port_group(self):
        self.assertRaises(NotImplementedError,
                          self.driver.prepare_port_group, None, None, None)

    def test_update_port_group(self):
        self.assertRaises(NotImplementedError,
                          self.driver.update_port_group, None, None, None)

    def test_get_vlanid_for_port_group(self):
        self.assertRaises(NotImplementedError,
                          self.driver.get_vlanid_for_port_group, None, None)

    def test_get_vm_ref_by_uuid(self):
        self.assertRaises(NotImplementedError,
                          self.driver.get_vm_ref_by_uuid, None)

    def test_wait_for_portgroup(self):
        self.assertRaises(NotImplementedError,
                          self.driver.wait_for_portgroup, None, None)

    def test_process_delete_vm(self):
        self.assertRaises(NotImplementedError,
                          self.driver.process_delete_vm, None)


class TestNetworkDriverCallback(base.TestCase):

    def setUp(self):
        super(TestNetworkDriverCallback, self).setUp()
        self.callback = driver.NetworkDriverCallback()

    def test_process_event(self):
        self.assertRaises(NotImplementedError,
                          self.callback.process_event, None)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from alembic import config as alembic_config

from neutron.db.migration import cli

CONF = cli.CONF


def get_alembic_config():
    script_location = 'networking_vsphere.db.migration:alembic_migrations'
    config = alembic_config.Config(os.path.join(os.path.dirname(__file__),
                                                'alembic.ini'))
    config.set_main_option('script_location', script_location)
    return config


def main():
    config = get_alembic_config()
    config.neutron_config = CONF
    CONF()
    CONF.command.func(config, CONF.command.name)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock
from oslo_config import cfg

from neutron.agent import securitygroups_rpc
from neutron.openstack.common import uuidutils

from networking_vsphere.agent import ovsvapp_sg_agent
from networking_vsphere.tests import base

cfg.CONF.import_group('AGENT', 'neutron.plugins.openvswitch.common.config')

fake_port = {'security_group_source_groups': 'abc',
             'mac_address': '00:11:22:33:44:55',
             'network_id': "netid",
             'id': "123",
             'security_groups': "abc",
             'segmentation_id': "100",
             "security_group_rules": [
                 {"direction": "ingress",
                  "protocol": "tcp",
                  "port_range_min": 2001,
                  "port_range_max": 2009,
                  "source_port_range_min": 67,
                  "source_port_range_max": 77,
                  "ethertype": "IPv4",
                  "source_ip_prefix": "150.1.1.0/22",
                  "dest_ip_prefix": "170.1.1.0/22"}]}


class FakePlugin(securitygroups_rpc.SecurityGroupServerRpcApi):
    def __init__(self, topic):
        self.topic = topic


class TestOVSVAppSecurityGroupAgent(base.TestCase):

    def setUp(self):
        super(TestOVSVAppSecurityGroupAgent, self).setUp()
        self.context = mock.Mock()
        self.plugin = FakePlugin('fake_topic')
        cfg.CONF.set_override('security_bridge_mapping',
                              "br-fake:fake_if", 'SECURITYGROUP')
        with contextlib.nested(
                mock.patch('networking_vsphere.drivers.'
                           'ovs_firewall.OVSFirewallDriver.'
                           'setup_base_flows'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'create'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'set_secure_mode'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'get_port_ofport',
                           return_value=5)):
            self.agent = ovsvapp_sg_agent.OVSVAppSecurityGroupAgent(
                self.context, self.plugin, True)
        self.LOG = ovsvapp_sg_agent.LOG

    def test_add_devices_to_filter(self):
        with mock.patch.object(self.agent.firewall,
                               'add_ports_to_filter') as mock_add:
            self.agent.add_devices_to_filter(['213421adsfs-123412asdfas',
                                              '213421adsfs-123412asdfas'])
            mock_add.assert_called_with(['213421adsfs-123412asdfas',
                                         '213421adsfs-123412asdfas'])

    def test_add_devices_to_filter_empty_input(self):
        with mock.patch.object(self.agent.firewall,
                               'add_ports_to_filter') as mock_add:
            self.agent.add_devices_to_filter([])
            self.assertFalse(mock_add.called)

    def test_ovsvapp_sg_update(self):
        ports = {"123": fake_port['security_group_rules']}
        self.agent.firewall.filtered_ports["123"] = fake_port
        with mock.patch.object(self.agent.firewall,
                               'prepare_port_filter') as mock_prepare:
            self.agent.ovsvapp_sg_update(ports)
            self.assertTrue(mock_prepare.called)
            mock_prepare.assert_called_with(ports["123"])

    def test_ovsvapp_sg_update_multiple_ports(self):
        ports = {"123": fake_port['security_group_rules'],
                 "456": fake_port['security_group_rules']}
        self.agent.firewall.filtered_ports["123"] = fake_port
        self.agent.firewall.filtered_ports["456"] = fake_port
        with mock.patch.object(self.agent.firewall,
                               'prepare_port_filter') as mock_prepare:
            self.agent.ovsvapp_sg_update(ports)
            self.assertTrue(mock_prepare.called)
            mock_prepare.assert_called_with(ports["123"])
            mock_prepare.assert_called_with(ports["456"])

    def test_ovsvapp_sg_update_invalid_port(self):
        ports = {"123": fake_port['security_group_rules']}
        self.agent.firewall.filtered_ports = {}
        with mock.patch.object(self.agent.firewall,
                               'prepare_port_filter') as mock_prepare:
            self.agent.ovsvapp_sg_update(ports)
            self.assertFalse(mock_prepare.called)

    def test_remove_device_filters(self):
        with mock.patch.object(self.agent.firewall,
                               'clean_port_filters') as mock_clean:
            self.agent.remove_device_filters("123")
            self.assertTrue(mock_clean.called)
            mock_clean.assert_called_with(["123"], True)

    def _get_fake_portids(self, count):
        ports = []
        i = 0
        while i < count:
            ports.append(uuidutils.generate_uuid())
            i += 1
        return ports

    def _get_fake_ports(self, ids):
        ports = {}
        for id in ids:
            port = {'id': id, 'security_group_rules': mock.Mock()}
            ports[id] = port
        return ports

    def test_prepare_firewall(self):
        port_ids = self._get_fake_portids(2)
        ret_val = self._get_fake_ports(port_ids)
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'prepare_port_filter')
        ) as (plugin_rpc, mock_prep):
            self.agent.prepare_firewall(set(port_ids))
            self.assertEqual(1, plugin_rpc.call_count)
            self.assertEqual(2, mock_prep.call_count)

    def test_prepare_firewall_many_ports(self):
        # Since we batch the ports before firing RPC to the neutron server
        # the number of RPCs should be lower. To be specific, we should be
        # firing one RPC call for every 10 ports.
        port_ids = self._get_fake_portids(25)
        ret_val = self._get_fake_ports(port_ids)
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'prepare_port_filter')
        ) as (plugin_rpc, mock_prep):
            self.agent.prepare_firewall(set(port_ids))
            self.assertEqual(3, plugin_rpc.call_count)
            self.assertEqual(25, mock_prep.call_count)

    def test_refresh_firewall_specific_ports(self):
        port_ids = self._get_fake_portids(2)
        ret_val = self._get_fake_ports(port_ids)
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'update_port_filter')
        ) as (plugin_rpc, mock_update):
            self.agent.refresh_firewall(set(port_ids))
            self.assertEqual(1, plugin_rpc.call_count)
            self.assertEqual(2, mock_update.call_count)

    def test_refresh_firewall_many_ports(self):
        # Since we batch the ports before firing RPC to the neutron server
        # the number of RPCs should be lower. To be specific, we should be
        # firing one RPC call for every 10 ports.
        port_ids = self._get_fake_portids(30)
        ret_val = self._get_fake_ports(port_ids)
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'update_port_filter')
        ) as (plugin_rpc, mock_update):
            self.agent.refresh_firewall(set(port_ids))
            self.assertEqual(3, plugin_rpc.call_count)
            self.assertEqual(30, mock_update.call_count)

    def test_refresh_firewall_no_input(self):
        port_ids = self._get_fake_portids(20)
        ret_val = self._get_fake_ports(port_ids)
        self.agent.firewall.filtered_ports = ret_val
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'update_port_filter')
        ) as (plugin_rpc, mock_update):
            self.agent.refresh_firewall()
            self.assertEqual(2, plugin_rpc.call_count)
            self.assertEqual(20, mock_update.call_count)

    def test_refresh_firewall_no_input_firewall_empty(self):
        port_ids = self._get_fake_portids(20)
        ret_val = self._get_fake_ports(port_ids)
        self.agent.firewall.filtered_ports = {}
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              'security_group_rules_for_devices',
                              return_value=ret_val),
            mock.patch.object(self.agent.firewall,
                              'update_port_filter')
        ) as (plugin_rpc, mock_update):
            self.agent.refresh_firewall()
            self.assertFalse(plugin_rpc.called)
            self.assertFalse(mock_update.called)

    def test_refresh_port_filters_case1(self):
        # Global refresh firewall is false devices_to_refilter has few ports
        # This case is where we need to update filters only the own ports
        own_host_ports = self._get_fake_portids(3)
        other_host_ports = self._get_fake_portids(6)
        self.agent.devices_to_refilter = set(own_host_ports)
        self.agent.global_refresh_firewall = False
        with contextlib.nested(
            mock.patch.object(self.agent.firewall, 'clean_port_filters'),
            mock.patch.object(self.agent, 'refresh_firewall'),
            mock.patch.object(self.agent, 'prepare_firewall')
        ) as (mock_clean, mock_refresh, mock_prepare):
            self.agent.refresh_port_filters(set(own_host_ports),
                                            set(other_host_ports))
        self.assertFalse(self.agent.devices_to_refilter)
        self.assertFalse(self.agent.global_refresh_firewall)
        mock_clean.assert_called_with(set([]))
        self.assertTrue(mock_refresh.called)
        self.assertFalse(mock_prepare.called)

    def test_refresh_port_filters_case2(self):
        # Global refresh firewall is false devices_to_refilter has few ports
        # This case is where we need to update filters only the other ports
        own_host_ports = self._get_fake_portids(3)
        other_host_ports = self._get_fake_portids(6)
        self.agent.devices_to_refilter = set(other_host_ports)
        self.agent.global_refresh_firewall = False
        with contextlib.nested(
            mock.patch.object(self.agent.firewall, 'clean_port_filters'),
            mock.patch.object(self.agent, 'refresh_firewall'),
            mock.patch.object(self.agent, 'prepare_firewall')
        ) as (mock_clean, mock_refresh, mock_prepare):
            self.agent.refresh_port_filters(set(own_host_ports),
                                            set(other_host_ports))
        self.assertFalse(self.agent.devices_to_refilter)
        self.assertFalse(self.agent.global_refresh_firewall)
        mock_clean.assert_called_with(set(other_host_ports))
        self.assertFalse(mock_refresh.called)
        self.assertTrue(mock_prepare.called)

    def test_refresh_port_filters_case3(self):
        # Global refresh firewall is false devices_to_refilter has few ports
        # This case is where we need to update filters for both own host and
        # other host ports
        own_host_ports = self._get_fake_portids(3)
        other_host_ports = self._get_fake_portids(6)
        ports_to_filter = [own_host_ports[2], other_host_ports[3],
                           other_host_ports[4]]
        self.agent.devices_to_refilter = set(ports_to_filter)
        self.agent.global_refresh_firewall = False
        with contextlib.nested(
            mock.patch.object(self.agent.firewall, 'clean_port_filters'),
            mock.patch.object(self.agent, 'refresh_firewall'),
            mock.patch.object(self.agent, 'prepare_firewall')
        ) as (mock_clean, mock_refresh, mock_prepare):
            self.agent.refresh_port_filters(set(own_host_ports),
                                            set(other_host_ports))
        self.assertFalse(self.agent.devices_to_refilter)
        self.assertFalse(self.agent.global_refresh_firewall)
        mock_clean.assert_called_with(set([other_host_ports[3],
                                           other_host_ports[4]]))
        self.assertTrue(mock_refresh.called)
        self.assertTrue(mock_prepare.called)

    def test_refresh_port_filters_case4(self):
        # Global refresh firewall is True devices_to_refilter has no ports
        own_host_ports = self._get_fake_portids(3)
        other_host_ports = self._get_fake_portids(6)
        self.agent.devices_to_refilter = set([])
        self.agent.global_refresh_firewall = True
        with contextlib.nested(
            mock.patch.object(self.agent.firewall, 'clean_port_filters'),
            mock.patch.object(self.agent, 'refresh_firewall'),
            mock.patch.object(self.agent, 'prepare_firewall')
        ) as (mock_clean, mock_refresh, mock_prepare):
            self.agent.refresh_port_filters(set(own_host_ports),
                                            set(other_host_ports))
        self.assertFalse(self.agent.global_refresh_firewall)
        mock_clean.assert_called_with(set(other_host_ports))
        self.assertTrue(mock_refresh.called)
        self.assertFalse(mock_prepare.called)

    def test_refresh_port_filters_case5(self):
        # Global refresh firewall is True devices_to_refilter has few ports
        # global_refresh_firewall should get priority over devices_to_refilter
        own_host_ports = self._get_fake_portids(3)
        other_host_ports = self._get_fake_portids(6)
        self.agent.devices_to_refilter = set(own_host_ports)
        self.agent.global_refresh_firewall = True
        with contextlib.nested(
            mock.patch.object(self.agent.firewall, 'clean_port_filters'),
            mock.patch.object(self.agent, 'refresh_firewall'),
            mock.patch.object(self.agent, 'prepare_firewall')
        ) as (mock_clean, mock_refresh, mock_prepare):
            self.agent.refresh_port_filters(set(own_host_ports),
                                            set(other_host_ports))
        self.assertFalse(self.agent.devices_to_refilter)
        self.assertFalse(self.agent.global_refresh_firewall)
        mock_clean.assert_called_with(set(other_host_ports))
        self.assertTrue(mock_refresh.called)
        self.assertFalse(mock_prepare.called)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Exception classes
"""


class VimException(Exception):

    """The VIM Exception class."""

    def __init__(self, exception_summary, excep):
        Exception.__init__(self)
        self.exception_summary = exception_summary
        self.exception_obj = excep

    def __str__(self):
        return self.exception_summary + str(self.exception_obj)


class SocketTimeoutException(VimException):

    """Socket Timeout Exception."""
    pass


class VimFaultException(Exception):

    """The VIM Fault exception class."""

    def __init__(self, fault_list, excep):
        Exception.__init__(self)
        self.fault_list = fault_list
        self.exception_obj = excep

    def __str__(self):
        return str(self.exception_obj)


class RunTimeError(Exception):

    """vCenter Run time error."""
    pass


# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid as uuid1


class NetworkConfig(object):

    def __init__(self, vlan):
        self.vlan = vlan


class Vlan(object):

    def __init__(self, vlanIds=[], operation_mode=None, vlan_type="Native"):
        self.vlan_type = vlan_type
        self.operation_mode = operation_mode
        self.vlanIds = vlanIds


class ResourceEntity(object):

    def __init__(self, key=None, uuid=None):
        super(ResourceEntity, self).__init__()
        if not uuid:
            uuid = uuid1.uuid1()
        self.uuid = str(uuid)
        self.key = key


class Host(ResourceEntity):

    def __init__(self, name=None, key=None):
        super(Host, self).__init__(key)
        self.name = name


class PhysicalNic(ResourceEntity):

    def __init__(self, name, mac_address, config, key=None):
        super(PhysicalNic, self).__init__(key)
        self.name = name
        self.mac_address = mac_address
        self.config = config


class VirtualSwitch(ResourceEntity):

    def __init__(self, name, pnics=None, networks=None, hosts=None, key=None):
        super(VirtualSwitch, self).__init__(key)
        self.name = name
        self.pnics = pnics or []
        self.networks = networks or []
        self.hosts = hosts or []


class Network(ResourceEntity):

    def __init__(self, name, network_type, config=None,
                 vswitches=None, ports=None, key=None):
        super(Network, self).__init__(key)
        self.name = name
        self.network_type = network_type
        self.config = config
        self.vswitches = vswitches or []
        self.ports = ports or []


class Port(ResourceEntity):

    def __init__(self, name=None, mac_address=None,
                 ipaddresses=None, vswitch_uuid=None,
                 vm_id=None, network_uuid=None, port_config=None,
                 port_status=None, key=None,
                 uuid=None):
        super(Port, self).__init__(key, uuid)
        self.name = name
        self.mac_address = mac_address
        self.ipaddresses = ipaddresses
        self.vswitch_uuid = vswitch_uuid
        self.vm_id = vm_id
        self.network_uuid = network_uuid
        self.port_config = port_config
        self.port_status = port_status


class VirtualNic(ResourceEntity):

    def __init__(self, mac_address, port_uuid,
                 vm_id, vm_name, nic_type, key=None):
        super(VirtualNic, self).__init__(key)
        self.mac_address = mac_address
        self.port_uuid = port_uuid
        self.vm_id = vm_id
        self.vm_name = vm_name
        self.nic_type = nic_type


class VirtualMachine(ResourceEntity):

    def __init__(self, name, vnics, uuid=None, key=None):
        if uuid:
            super(VirtualMachine, self).__init__(key, uuid)
        else:
            super(VirtualMachine, self).__init__(key)
        self.name = name
        self.vnics = vnics


class Event(object):

    def __init__(self, event_type, src_obj, changes, host_name,
                 cluster_name, cluster_id, host_changed):
        self.event_type = event_type
        self.src_obj = src_obj
        self.changes = changes
        self.host_name = host_name
        self.cluster_name = cluster_name
        self.cluster_id = cluster_id
        self.host_changed = host_changed



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_vsphere.drivers import driver as network_driver
from networking_vsphere.drivers import driver as network_driver_callback


class FakeInvalidDriver(object):
    pass


class FakeNetworkDriver(network_driver.NetworkDriver):
    pass


class MockNetworkDriver(network_driver.NetworkDriver):

    def __init__(self):
        super(MockNetworkDriver, self).__init__()
        self.methods = {}

    def create_network(self, network, virtual_switch):
        kwargs = {}
        kwargs["network"] = network
        kwargs["virtual_switch"] = virtual_switch
        self.methods["create_network"] = kwargs

    def create_port(self, network, port, virtual_nic):
        kwargs = {}
        kwargs["network"] = network
        kwargs["port"] = port
        kwargs["virtual_nic"] = virtual_nic
        self.methods["create_port"] = kwargs

    def update_port(self, network=None, port=None, virtual_nic=None):
        kwargs = {}
        kwargs["network"] = network
        kwargs["port"] = port
        kwargs["virtual_nic"] = virtual_nic
        self.methods["update_port"] = kwargs

    def prepare_port_group(self, network, port, virtual_nic):
        kwargs = {}
        kwargs["network"] = network
        kwargs["port"] = port
        kwargs["virtual_nic"] = virtual_nic
        self.methods["prepare_port_group"] = kwargs

    def update_port_group(self, network=None, port=None, virtual_nic=None):
        kwargs = {}
        kwargs["network"] = network
        kwargs["port"] = port
        kwargs["virtual_nic"] = virtual_nic
        self.methods["update_port_group"] = kwargs

    def delete_network(self, network, virtual_switch=None):
        kwargs = {}
        kwargs["network"] = network
        kwargs["virtual_switch"] = virtual_switch
        self.methods["delete_network"] = kwargs

    def post_delete_vm(self, vm):
        kwargs = {}
        kwargs["vm"] = vm
        self.methods["post_delete_vm"] = kwargs

    def process_delete_vm(self, vm):
        kwargs = {}
        kwargs["vm"] = vm
        self.methods["process_delete_vm"] = kwargs

    def get_vlanid_for_port_group(self, dvs_name, pg_name):
        kwargs = {}
        kwargs["dvs_name"] = dvs_name
        kwargs["pg_name"] = pg_name
        self.methods["get_vlanid_for_port_group"] = kwargs

    def get_vm_ref_by_uuid(self, vm_uuid):
        kwargs = {}
        kwargs["vm_uuid"] = vm_uuid
        self.methods["get_vm_ref_by_uuid"] = kwargs

    def wait_for_portgroup(self, vm_ref, pg_name):
        kwargs = {}
        kwargs["vm_ref"] = vm_ref
        kwargs["pg_name"] = pg_name
        self.methods["wait_for_portgroup"] = kwargs

    def monitor_events(self):
        self.methods["monitor_events"] = {}

    def is_connected(self):
        self.methods["is_connected"] = {}

    def stop(self):
        self.methods["stop"] = {}

    def pause(self):
        self.methods["pause"] = {}

    def reset(self):
        self.methods = {}


class MockCallback(network_driver_callback.NetworkDriverCallback):

    def __init__(self):
        self.events = []

    def process_event(self, event):
        self.events.append(event)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

import netaddr
from oslo_config import cfg
from oslo_log import log

from neutron.agent.common import ovs_lib
from neutron.agent import firewall
from neutron.common import constants

from networking_vsphere.common import constants as ovsvapp_const

LOG = log.getLogger(__name__)

PROTOCOLS = {constants.PROTO_NAME_TCP: constants.PROTO_NAME_TCP,
             constants.PROTO_NUM_TCP: constants.PROTO_NAME_TCP,
             constants.PROTO_NAME_UDP: constants.PROTO_NAME_UDP,
             constants.PROTO_NUM_UDP: constants.PROTO_NAME_UDP,
             constants.PROTO_NAME_ICMP: constants.PROTO_NAME_ICMP,
             constants.PROTO_NUM_ICMP: constants.PROTO_NAME_ICMP,
             constants.PROTO_NAME_ICMP_V6: constants.PROTO_NAME_ICMP_V6,
             constants.PROTO_NUM_ICMP_V6: constants.PROTO_NAME_ICMP_V6}

ETHERTYPE = {constants.IPv4: "ip",
             constants.IPv6: "ipv6"}

INGRESS_DIRECTION = 'ingress'
EGRESS_DIRECTION = 'egress'

sg_conf = cfg.CONF.SECURITYGROUP

PORT_KEYS = ['security_group_source_groups',
             'mac_address',
             'network_id',
             'id',
             'security_groups',
             'lvid']


class OVSFirewallDriver(firewall.FirewallDriver):
    """Driver which enforces security groups through OVS flows."""

    def __init__(self):
        self.filtered_ports = {}
        if sg_conf.security_bridge_mapping is None:
            LOG.debug("Security_bridge_mapping not configured.")
            return
        secbr_list = (sg_conf.security_bridge_mapping).split(':')
        secbr_name = secbr_list[0]
        secbr_phyname = secbr_list[1]
        self.sg_br = ovs_lib.OVSBridge(secbr_name)
        self.phy_ofport = self.sg_br.get_port_ofport(secbr_phyname)
        self.patch_ofport = self.sg_br.get_port_ofport(
            ovsvapp_const.SEC_TO_INT_PATCH)
        self._defer_apply = False
        if not cfg.CONF.OVSVAPP.agent_maintenance:
            self.setup_base_flows()

    @property
    def ports(self):
        return self.filtered_ports

    def _get_compact_port(self, port):
        if 'lvid' not in port:
            if port['id'] in self.filtered_ports:
                old_port = self.filtered_ports[port['id']]
                port['lvid'] = old_port['lvid']
        new_port = {}
        new_port['device'] = port['id']
        for key in PORT_KEYS:
            if key in port:
                new_port[key] = port[key]
        return new_port

    def _add_ovs_flow(self, sg_br, pri, table_id, action,
                      protocol=None, dl_dest=None,
                      tcp_flag=None, icmp_req_type=None):
        """Helper method for adding OVS flows.

        Method which will help add an openflow rule with the given
        priority and action in the specified table.
        """
        if protocol:
            sg_br.add_flow(table=table_id, priority=pri,
                           proto=protocol, actions=action)
        elif dl_dest:
            sg_br.add_flow(table=table_id, priority=pri,
                           dl_dst=dl_dest, actions=action)
        elif tcp_flag:
            sg_br.add_flow(table=table_id, priority=pri,
                           proto=constants.PROTO_NAME_TCP,
                           tcp_flags=tcp_flag, actions=action)
        elif icmp_req_type:
            sg_br.add_flow(table=table_id, priority=pri,
                           proto=constants.PROTO_NAME_ICMP,
                           icmp_type=icmp_req_type,
                           actions=action)
        else:
            sg_br.add_flow(table=table_id, priority=pri, actions=action)

    def _add_icmp_learn_flow(self, sec_br, reqType, resType,
                             pri=ovsvapp_const.SG_TP_PRI):
        sec_br.add_flow(priority=pri,
                        table=ovsvapp_const.SG_ICMP_TABLE_ID,
                        proto=constants.PROTO_NAME_ICMP,
                        icmp_type=reqType,
                        actions="learn(%s)" %
                        self._get_icmp_learn_flow(resType))

    def _get_icmp_learn_flow(self, resType):
        if resType is ovsvapp_const.ICMP_DEST_UNREACH:
            ip_str = ""
        else:
            ip_str = "NXM_OF_IP_SRC[]=NXM_OF_IP_DST[],"
        return ("table=%s,"
                "priority=%s,"
                "idle_timeout=30,"
                "dl_type=0x0800,"
                "NXM_OF_VLAN_TCI[0..11],"
                "nw_proto=%s,"
                "icmp_type=%s,"
                "%s"
                "NXM_OF_IP_DST[]=NXM_OF_IP_SRC[],"
                "output:NXM_OF_IN_PORT[]" %
                (ovsvapp_const.SG_LEARN_TABLE_ID,
                 ovsvapp_const.SG_TP_PRI,
                 constants.PROTO_NUM_ICMP,
                 resType, ip_str))

    def _setup_icmp_learn_flows(self, sec_br):
        # ICMP Learned flows
        self._add_icmp_learn_flow(sec_br, ovsvapp_const.ICMP_ECHO_REQ,
                                  ovsvapp_const.ICMP_ECHO_REP)
        self._add_icmp_learn_flow(sec_br, ovsvapp_const.ICMP_TS_REQ,
                                  ovsvapp_const.ICMP_TS_REP)
        self._add_icmp_learn_flow(sec_br, ovsvapp_const.ICMP_INFO_REQ,
                                  ovsvapp_const.ICMP_INFO_REP)
        self._add_icmp_learn_flow(sec_br, ovsvapp_const.ICMP_AM_REQ,
                                  ovsvapp_const.ICMP_AM_REP)

    def _setup_learning_flows(self, sec_br):
        """Helper method for adding learning flows.

        Method which will help setup the base learning flows at
        the start of the agent.
        These flows are populated in specific tables for
        TCP/UDP/ICMP.
        """
        # First we chain the tables.
        self._add_ovs_flow(sec_br, ovsvapp_const.SG_DEFAULT_PRI,
                           ovsvapp_const.SG_ICMP_TABLE_ID, "drop")
        # If DMAC is bcast or mcast, don't learn.
        self._add_ovs_flow(sec_br, ovsvapp_const.SG_DROP_HIGH_PRI,
                           ovsvapp_const.SG_IP_TABLE_ID, "drop",
                           dl_dest="01:00:00:00:00:00/01:00:00:00:00:00")

        # Now we add learning flows one for TCP and another for UDP.
        learned_tcp_flow = ("table=%s,"
                            "priority=%s,"
                            "fin_idle_timeout=1,"
                            "idle_timeout=7200,"
                            "nw_proto=%s,"
                            "NXM_OF_ETH_SRC[]=NXM_OF_ETH_DST[],"
                            "NXM_OF_ETH_DST[]=NXM_OF_ETH_SRC[],"
                            "dl_type=0x0800,"
                            "NXM_OF_VLAN_TCI[0..11],"
                            "NXM_OF_IP_SRC[]=NXM_OF_IP_DST[],"
                            "NXM_OF_IP_DST[]=NXM_OF_IP_SRC[],"
                            "NXM_OF_TCP_SRC[]=NXM_OF_TCP_DST[],"
                            "NXM_OF_TCP_DST[]=NXM_OF_TCP_SRC[],"
                            "output:NXM_OF_IN_PORT[]" %
                            (ovsvapp_const.SG_LEARN_TABLE_ID,
                             ovsvapp_const.SG_TP_PRI,
                             constants.PROTO_NUM_TCP))
        self._add_ovs_flow(sec_br, ovsvapp_const.SG_TP_PRI,
                           ovsvapp_const.SG_TCP_TABLE_ID,
                           "learn(%s)" % learned_tcp_flow,
                           protocol=constants.PROTO_NAME_TCP)

        learned_udp_flow = ("table=%s,"
                            "priority=%s,"
                            "idle_timeout=300,"
                            "nw_proto=%s,"
                            "NXM_OF_ETH_SRC[]=NXM_OF_ETH_DST[],"
                            "NXM_OF_ETH_DST[]=NXM_OF_ETH_SRC[],"
                            "dl_type=0x0800,"
                            "NXM_OF_VLAN_TCI[0..11],"
                            "NXM_OF_IP_SRC[]=NXM_OF_IP_DST[],"
                            "NXM_OF_IP_DST[]=NXM_OF_IP_SRC[],"
                            "NXM_OF_UDP_SRC[]=NXM_OF_UDP_DST[],"
                            "NXM_OF_UDP_DST[]=NXM_OF_UDP_SRC[],"
                            "output:NXM_OF_IN_PORT[]" %
                            (ovsvapp_const.SG_LEARN_TABLE_ID,
                             ovsvapp_const.SG_TP_PRI,
                             constants.PROTO_NUM_UDP))
        self._add_ovs_flow(sec_br, ovsvapp_const.SG_TP_PRI,
                           ovsvapp_const.SG_UDP_TABLE_ID,
                           "learn(%s)" % learned_udp_flow,
                           protocol=constants.PROTO_NAME_UDP)
        # Now setup the ICMP learn flows.
        self._setup_icmp_learn_flows(sec_br)

    def setup_base_flows(self):
        """Method for configuring the default flows in OVS bridge.

        Method which will help setup the base flows at the start
        of the agent.
        """
        try:
            sec_br = self.sg_br
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_DEFAULT_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "resubmit(,%s)" %
                               ovsvapp_const.SG_LEARN_TABLE_ID)
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_DROPALL_PRI,
                               ovsvapp_const.SG_LEARN_TABLE_ID, "drop")
            # Allow all ARP, parity with iptables.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_RULES_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "normal", protocol="arp")
            # Allow all RARP, parity with iptables.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_RULES_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "normal", protocol="rarp")
            # Rule to allow VMs to send DHCP requests (udp).
            sec_br.add_flow(priority=ovsvapp_const.SG_RULES_PRI,
                            table=ovsvapp_const.SG_DEFAULT_TABLE_ID,
                            proto="udp", tp_src="68", tp_dst="67",
                            actions="normal")
            # Always allow ICMP DestUnreach.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_TP_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "normal", icmp_req_type=ovsvapp_const.
                               ICMP_DEST_UNREACH)

            # Always allow ICMP TTL Exceeded.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_TP_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID, "normal",
                               icmp_req_type=ovsvapp_const.
                               ICMP_TIME_EXCEEDED)

            self._setup_learning_flows(sec_br)

            # Always resubmit FIN pkts to learn table.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_TCP_FLAG_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "resubmit(,%s),normal" %
                               ovsvapp_const.SG_LEARN_TABLE_ID,
                               tcp_flag='+fin')
            # Always resubmit RST pkts to learn table.
            self._add_ovs_flow(sec_br, ovsvapp_const.SG_TCP_FLAG_PRI,
                               ovsvapp_const.SG_DEFAULT_TABLE_ID,
                               "resubmit(,%s),normal" %
                               ovsvapp_const.SG_LEARN_TABLE_ID,
                               tcp_flag='+rst')
        except Exception:
            LOG.exception(_("Unable to add base flows."))

    def add_ports_to_filter(self, ports):
        for port in ports:
            LOG.debug("OVSF Adding port: %s to filter.", port)
            self.filtered_ports[port['id']] = self._get_compact_port(port)

    def _get_port_vlan(self, port_id):
        if port_id:
            port = self.filtered_ports.get(port_id)
            LOG.debug("Filtered port: %s.", port)
            if port:
                return port['lvid']

    def _setup_aap_flows(self, sec_br, port):
        """Method to help setup rules for allowed address pairs."""
        vlan = self._get_port_vlan(port['id'])
        if not vlan:
            LOG.error(_("Missing VLAN information for port: %s."), port['id'])
            return
        if isinstance(port.get('allowed_address_pairs'), list):
            for addr_pair in port['allowed_address_pairs']:
                if netaddr.IPNetwork(addr_pair["ip_address"]).version == 4:
                    ap_proto = "ip"
                else:
                    ap_proto = "ipv6"
                sec_br.add_flow(priority=ovsvapp_const.SG_RULES_PRI,
                                table=ovsvapp_const.SG_DEFAULT_TABLE_ID,
                                cookie=self.get_cookie(port['id']),
                                dl_dst=port["mac_address"],
                                in_port=self.patch_ofport,
                                dl_src=addr_pair["mac_address"],
                                dl_vlan=vlan,
                                proto=ap_proto,
                                nw_src=addr_pair["ip_address"],
                                actions="resubmit(,%s),output:%s" %
                                (ovsvapp_const.SG_IP_TABLE_ID,
                                 self.phy_ofport))

    def _get_net_prefix_len(self, ip_prefix):
        if ip_prefix:
            return netaddr.IPNetwork(ip_prefix).prefixlen
        else:
            return 0

    def _get_protocol(self, ethertype, proto):
        if ethertype == constants.IPv6 and (
            proto == constants.PROTO_NAME_ICMP):
            return ['icmp6']
        elif proto is not None:
            protocol = PROTOCOLS.get(proto)
            if protocol is None:
                return [ETHERTYPE.get(ethertype), proto]
            else:
                return [protocol]
        elif ethertype == constants.IPv4:
            return ['ip']
        elif ethertype == constants.IPv6:
            return ['ipv6']

    def _add_flow_with_range(self, sec_br, flow,
                             pr_min=None, pr_max=None,
                             spr_min=None, spr_max=None):
        if pr_min is None and pr_max is None:
            pr_min = -1
            pr_max = -1
        if spr_min is None and spr_max is None:
            spr_min = -1
            spr_max = -1

        for dport, sport in itertools.product(xrange(pr_min, pr_max + 1),
                                              xrange(spr_min, spr_max + 1)):
            if dport >= 0:
                flow["tp_dst"] = dport
            if sport >= 0:
                flow["tp_src"] = sport
            sec_br.add_flow(**flow)

    def _add_flows(self, sec_br, port, rules=None):
        egress_action = 'normal'
        ingress_action = 'output:%s' % self.phy_ofport

        if not rules:
            rules = port["security_group_rules"]

        vlan = self._get_port_vlan(port['id'])
        if not vlan:
            LOG.error(_('Missing VLAN for port: %s.'), port['id'])
            return
        for rule in rules:
            direction = rule.get('direction')
            proto = rule.get('protocol')
            pr_min = rule.get('port_range_min')
            pr_max = rule.get('port_range_max')
            spr_min = rule.get('source_port_range_min')
            spr_max = rule.get('source_port_range_max')
            ethertype = rule.get('ethertype')
            src_ip_prefix = rule.get('source_ip_prefix')
            dest_ip_prefix = rule.get('dest_ip_prefix')
            flow = dict(priority=ovsvapp_const.SG_RULES_PRI)
            flow["table"] = ovsvapp_const.SG_DEFAULT_TABLE_ID
            # Using port id as cookie.
            flow["cookie"] = self.get_cookie(port['id'])
            flow["dl_vlan"] = vlan
            # Fill the src and dest IPs match params.
            src_ip_prefixlen = self._get_net_prefix_len(src_ip_prefix)
            if src_ip_prefixlen > 0:
                flow["nw_src"] = src_ip_prefix
            dest_ip_prefixlen = self._get_net_prefix_len(dest_ip_prefix)
            if dest_ip_prefixlen > 0:
                flow["nw_dst"] = dest_ip_prefix
            # Fill the protocol related  match params.
            protocols = self._get_protocol(ethertype, proto)
            protocol = protocols[0]
            flow["proto"] = protocol
            if len(protocols) > 1:
                flow["nw_proto"] = protocols[1]
            # set source and destination params and action for the flow.
            if direction == INGRESS_DIRECTION:
                flow["dl_dst"] = port["mac_address"]
                flow["in_port"] = self.patch_ofport
                action = ingress_action
            elif direction == EGRESS_DIRECTION:
                flow["dl_src"] = port["mac_address"]
                flow["in_port"] = self.phy_ofport
                action = egress_action
            tcp_udp = set([constants.PROTO_NAME_TCP,
                           constants.PROTO_NAME_UDP])
            table_id = None
            if protocol in tcp_udp:
                flow["priority"] = ovsvapp_const.SG_TP_PRI
                if protocol == constants.PROTO_NAME_TCP:
                    table_id = ovsvapp_const.SG_TCP_TABLE_ID
                else:
                    table_id = ovsvapp_const.SG_UDP_TABLE_ID
                flow["actions"] = ("resubmit(,%s),%s" % (table_id, action))
                self._add_flow_with_range(sec_br, flow, pr_min, pr_max,
                                          spr_min, spr_max)
                # Since we added the required flows in the above method
                # we just proceed to the next sg rule.
                continue
            elif protocol == constants.PROTO_NAME_ICMP:
                flow["priority"] = ovsvapp_const.SG_TP_PRI
                if pr_min is not None:
                    flow["icmp_type"] = pr_min
                if pr_max is not None:
                    flow["icmp_code"] = pr_max
                table_id = ovsvapp_const.SG_ICMP_TABLE_ID
            else:
                table_id = ovsvapp_const.SG_IP_TABLE_ID

            flow["actions"] = ("resubmit(,%s),%s" % (table_id, action))
            sec_br.add_flow(**flow)
            LOG.debug("OVSF adding flow: %s", flow)

    def prepare_port_filter(self, port):
        """Method to add OVS rules for a newly created VM port."""
        LOG.debug("OVSF Preparing port %s filter.", port['id'])
        try:
            with self.sg_br.deferred(full_ordered=True, order=(
                'del', 'mod', 'add')) as deferred_br:
                self._setup_aap_flows(deferred_br, port)
                self._add_flows(deferred_br, port)
            self.filtered_ports[port['id']] = self._get_compact_port(port)
        except Exception:
            LOG.exception(_("Unable to add flows for %s."), port['id'])

    def _remove_flows(self, sec_br, port_id):
        """Remove all flows for a port."""
        LOG.debug("OVSF Removing flows start for port: %s.", port_id)
        try:
            sec_br.delete_flows(cookie="%s/-1" %
                                self.get_cookie(port_id))
            port = self.filtered_ports.get(port_id)
            vlan = self._get_port_vlan(port_id)
            if 'mac_address' not in port or not vlan:
                LOG.debug("Invalid mac address or vlan for port "
                          "%s. Returning from _remove_flows.", port_id)
                return
            sec_br.delete_flows(table=ovsvapp_const.SG_LEARN_TABLE_ID,
                                dl_src=port['mac_address'],
                                vlan_tci="0x%04x/0x0fff" % vlan)
            sec_br.delete_flows(table=ovsvapp_const.SG_LEARN_TABLE_ID,
                                dl_dst=port['mac_address'],
                                vlan_tci="0x%04x/0x0fff" % vlan)
            sec_br.delete_flows(table=ovsvapp_const.SG_DEFAULT_TABLE_ID,
                                dl_src=port['mac_address'],
                                vlan_tci="0x%04x/0x0fff" % vlan)
            sec_br.delete_flows(table=ovsvapp_const.SG_DEFAULT_TABLE_ID,
                                dl_dst=port['mac_address'],
                                vlan_tci="0x%04x/0x0fff" % vlan)
        except Exception:
            LOG.exception(_("Unable to remove flows %s."), port['id'])

    def clean_port_filters(self, ports, remove_port=False):
        """Method to remove OVS rules for an existing VM port."""
        LOG.debug("OVSF Cleaning filters for  %s ports.", len(ports))
        if not ports:
            return
        with self.sg_br.deferred() as deferred_sec_br:
            for port_id in ports:
                try:
                    if not self.filtered_ports.get(port_id):
                        LOG.debug("Attempted to remove port filter "
                                  "which is not in filtered %s.", port_id)
                        continue
                    self._remove_flows(deferred_sec_br, port_id)
                    if remove_port:
                        self.filtered_ports.pop(port_id, None)
                except Exception:
                    LOG.exception(_("Unable to delete flows for %s."), port_id)

    def update_port_filter(self, port):
        """Method to update OVS rules for an existing VM port."""
        LOG.debug("OVSF Updating port: %s filter.", port['id'])
        if port['id'] not in self.filtered_ports:
            LOG.warn(_("Attempted to update port filter which is not "
                     "filtered %s."), port['id'])
            return
        try:
            with self.sg_br.deferred(full_ordered=True, order=(
                'del', 'mod', 'add')) as deferred_br:
                self._remove_flows(deferred_br, port['id'])
                self._setup_aap_flows(deferred_br, port)
                self._add_flows(deferred_br, port)
            self.filtered_ports[port['id']] = self._get_compact_port(port)
        except Exception:
            LOG.exception(_("Unable to update flows for %s."), port['id'])

    def filter_defer_apply_on(self):
        if not self._defer_apply:
            self._defer_apply = True

    def filter_defer_apply_off(self):
        if self._defer_apply:
            self._defer_apply = False

    def get_cookie(self, port_id):
        return ("0x%x" % (hash(port_id) & 0xffffffffffffffff))



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Table for ovsvapp cluster vni allocations.

Revision ID: 3f77522fea53
Revises: start_ovsvapp_migration
Create Date: 2015-03-25 02:40:53.207016
"""

# revision identifiers, used by Alembic.
revision = '3f77522fea53'
down_revision = 'start_ovsvapp_migration'

from alembic import op
import sqlalchemy as sa


def downgrade():
    op.drop_table('ovsvapp_cluster_vni_allocations')


def upgrade():
    op.create_table('ovsvapp_cluster_vni_allocations',
                    sa.Column('vcenter_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('cluster_id', sa.String(length=255),
                              nullable=False),
                    sa.Column('lvid', sa.Integer(), nullable=False,
                              autoincrement=False),
                    sa.Column('network_id', sa.String(length=36)),
                    sa.Column('allocated', sa.Boolean(),
                              server_default=sa.sql.false(), nullable=False),
                    sa.Column('network_port_count', sa.Integer(),
                              server_default='0', nullable=False),
                    sa.PrimaryKeyConstraint('vcenter_id', 'cluster_id',
                                            'lvid')
                    )



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import sqlalchemy.orm.exc as sa_exc

from neutron.db import api as db_api

from networking_vsphere.db import ovsvapp_models as models

LOG = log.getLogger(__name__)


def _generate_vcenter_cluster_allocations(session, vcenter, cluster):
    bulk_size = 100
    allocs = []
    lvid_min = 1
    lvid_max = 4095

    for lvid in xrange(lvid_min, lvid_max):
        allocs.append({'vcenter_id': vcenter,
                       'cluster_id': cluster,
                       'lvid': lvid})
    if allocs:
        chuncked_allocs = (allocs[i:i + bulk_size] for i in
                           range(0, len(allocs), bulk_size))
        for bulk in chuncked_allocs:
            session.execute(models.ClusterVNIAllocations.
                            __table__.insert(), bulk)
    LOG.info(_("Finished initializing local vlans for cluster %(cluster)s "
               "of vCenter %(vcenter)s."), {'cluster': cluster,
                                            'vcenter': vcenter})


def _initialize_lvids_for_cluster(port_info):
    vcenter = port_info['vcenter_id']
    cluster = port_info['cluster_id']
    session = db_api.get_session()
    with session.begin(subtransactions=True):
        try:
            (session.query(models.ClusterVNIAllocations).
             with_lockmode('update')).all()
            query = session.query(models.ClusterVNIAllocations)
            existing_allocations = query.filter(
                models.ClusterVNIAllocations.vcenter_id == vcenter,
                models.ClusterVNIAllocations.cluster_id == cluster
            ).all()
            if not existing_allocations:
                _generate_vcenter_cluster_allocations(
                    session, vcenter, cluster)
            return True
        except Exception:
            LOG.exception(_("Exception while initializing VNI allocations for "
                            "clusters %(cluster)s of vCenter %(vcenter)s."),
                          {'cluster': cluster,
                           'vcenter': vcenter})
            return False


def _try_to_obtain_local_vlan(session, port_info, assign):
    lvid = None
    res_keys = ['vcenter_id', 'cluster_id', 'network_id']
    res = dict((k, port_info[k]) for k in res_keys)
    try:
        allocation = (session.query(models.ClusterVNIAllocations).filter(
            models.ClusterVNIAllocations.vcenter_id == res['vcenter_id'],
            models.ClusterVNIAllocations.cluster_id == res['cluster_id'],
            models.ClusterVNIAllocations.network_id == res['network_id']
        ).one())
        lvid = allocation.lvid
        if assign:
            count = allocation.network_port_count + 1
            allocation.update({'network_port_count': count})
    except sa_exc.NoResultFound:
        if not assign:
            raise Exception()
        try:
            allocation = session.query(models.ClusterVNIAllocations).filter(
                models.ClusterVNIAllocations.vcenter_id == res['vcenter_id'],
                models.ClusterVNIAllocations.cluster_id == res['cluster_id'],
                models.ClusterVNIAllocations.allocated == 0
            ).first()
            if allocation:
                lvid = allocation.lvid
                allocation.update({'network_id': res['network_id'],
                                   'allocated': True,
                                   'network_port_count': 1})
                LOG.info(_("Assigned local vlan %(lvid)s for the network "
                           "%(network)s on the cluster %(cluster)s."),
                         {'network': port_info['network_id'],
                          'cluster': port_info['cluster_id'],
                          'lvid': lvid})
            else:
                LOG.error(_("All available VLANs are used up in the cluster "
                            "%(cluster)s of vCenter %(vcenter)s."),
                          {'cluster': res['cluster_id'],
                           'vcenter': res['vcenter_id']})
        except Exception as e:
            LOG.exception(_("Unable to obtain local vlan id %s."), e)
    return lvid


def get_local_vlan(port_info, assign=True):
    lvid = None
    session = db_api.get_session()
    res_keys = ['vcenter_id', 'cluster_id', 'network_id']
    res = dict((k, port_info[k]) for k in res_keys)
    with session.begin(subtransactions=True):
        try:
            if not assign:
                lvid = _try_to_obtain_local_vlan(session, port_info, assign)
                return lvid
            query = session.query(models.ClusterVNIAllocations)
            # Lock all the rows in the table corresponding to the vCenter
            # and cluster.
            cluster_rows = query.filter(
                (models.ClusterVNIAllocations.vcenter_id == res['vcenter_id']),
                (models.ClusterVNIAllocations.cluster_id == res['cluster_id'])
            ).with_lockmode('update').all()
            if cluster_rows:
                lvid = _try_to_obtain_local_vlan(session, port_info, assign)
                return lvid
            else:
                LOG.info(_("Local VLAN rows not provisioned for the "
                           "cluster %(cluster)s of vCenter %(vcenter)s. "
                           "Going to provision."),
                         {'cluster': res['cluster_id'],
                          'vcenter': res['vcenter_id']})
        except Exception:
            LOG.exception(_("Error retrieving a local vlan for network "
                            "%(network)s for %(port)s."),
                          {'network': port_info['network_id'],
                           'port': port_info['port_id']})
            return
    status = _initialize_lvids_for_cluster(res)
    if status:
        with session.begin(subtransactions=True):
            lvid = _try_to_obtain_local_vlan(session, port_info, assign)
    else:
        LOG.error(_("Local VLAN rows not provisioned for the "
                    "cluster %(cluster)s of vCenter %(vcenter)s."),
                  {'cluster': res['cluster_id'],
                   'vcenter': res['vcenter_id']})
    return lvid


def check_to_reclaim_local_vlan(port_info):
    lvid = -1
    session = db_api.get_session()
    with session.begin(subtransactions=True):
        res_keys = ['vcenter_id', 'cluster_id', 'network_id']
        res = dict((k, port_info[k]) for k in res_keys)
        try:
            query = session.query(models.ClusterVNIAllocations)
            allocation = (query.filter(
                models.ClusterVNIAllocations.vcenter_id == res['vcenter_id'],
                models.ClusterVNIAllocations.cluster_id == res['cluster_id'],
                models.ClusterVNIAllocations.network_id == res['network_id']
            ).with_lockmode('update').one())
            count = allocation.network_port_count
            if count >= 1:
                count -= 1
                allocation.update({'network_port_count': count})
            if count == 0:
                lvid = allocation.lvid
        except sa_exc.NoResultFound:
            # Nothing to do, may be another controller cleared the record
            # We will just log and return back status as False.
            LOG.debug("Network %(network)s is already de-allocated for "
                      "cluster %(cluster)s.",
                      {'network': port_info['network_id'],
                       'cluster': port_info['cluster_id']})
    return lvid


def release_local_vlan(net_info):
    session = db_api.get_session()
    with session.begin(subtransactions=True):
        res_keys = ['vcenter_id', 'cluster_id', 'network_id']
        res = dict((k, net_info[k]) for k in res_keys)
        try:
            query = session.query(models.ClusterVNIAllocations)
            allocation = (query.filter(
                models.ClusterVNIAllocations.vcenter_id == res['vcenter_id'],
                models.ClusterVNIAllocations.cluster_id == res['cluster_id'],
                models.ClusterVNIAllocations.network_id == res['network_id']
            ).with_lockmode('update').one())
            if allocation.network_port_count == 0:
                allocation.update({'network_id': None,
                                   'allocated': False,
                                   'network_port_count': 0})
            else:
                LOG.info(_("Unable to release local vlan for network_id %s "
                           "because ports are available on network."),
                         res['network_id'])
        except sa_exc.NoResultFound:
            # Nothing to do, may be another controller cleared the record
            # We will just log and return.
            LOG.error(_("Network %(network)s is already de-allocated for "
                        "cluster %(cluster)s."),
                      {'network': net_info['network_id'],
                       'cluster': net_info['cluster_id']})


def get_stale_local_vlans_for_network(network_id):
    session = db_api.get_session()
    vcenter_clusters = None
    with session.begin(subtransactions=True):
        try:
            query = session.query(models.ClusterVNIAllocations)
            allocations = (query.filter(
                models.ClusterVNIAllocations.network_id == network_id
            ).all())
            if allocations:
                vcenter_clusters = []
                for alloc in allocations:
                    vcenter_clusters.append((alloc.vcenter_id,
                                             alloc.cluster_id,
                                             alloc.lvid))
        except Exception:
            # Nothing to do, port-deletions have properly cleaned up
            # the records. We will just log and return back empty list.
            LOG.debug("Network %s is already cleaned up from "
                      "VNI allocations table.", network_id)
    return vcenter_clusters



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
'''
Implements methods defined in NetworkDriver and VCNetworkDriver
and supports VMware Distributed Virtual Switch.
'''

from oslo_log import log

from networking_vsphere.common import constants
from networking_vsphere.common import error
from networking_vsphere.common import utils
from networking_vsphere.drivers import vc_driver
from networking_vsphere.utils import cache
from networking_vsphere.utils import common_util
from networking_vsphere.utils import error_util
from networking_vsphere.utils import network_util
from networking_vsphere.utils import resource_util
from networking_vsphere.utils import vim_util

LOG = log.getLogger(__name__)


class DvsNetworkDriver(vc_driver.VCNetworkDriver):

    def __init__(self):
        vc_driver.VCNetworkDriver.__init__(self)

    def get_unused_portgroups(self, switch):
        return network_util.get_unused_portgroup_names(self.session, switch)

    def delete_portgroup(self, switch, pg):
        network_util.delete_port_group(self.session,
                                       dvs_name=switch,
                                       pg_name=pg)

    def is_valid_switch(self, cluster_mor, switch):
        return network_util.is_valid_dvswitch(self.session,
                                              cluster_mor,
                                              switch)

    def _register_vm_for_updates(self, vm_mor, collector):
        vm_properties = ['name',
                         'config.hardware.device']
        propertyDict = {"VirtualMachine": vm_properties}
        property_filter_spec = self.session._call_method(
            vim_util, "get_property_filter_specs",
            propertyDict, [vm_mor])
        self.session._call_method(vim_util,
                                  "create_filter",
                                  property_filter_spec,
                                  collector)

    # TODO(romilg): Split into helper method
    # say _wait_for_vm_to_attach_to_portgroup().
    def _wait_for_port_update_on_vm(self, vm_mor, pgmor):
        property_collector = None
        try:
            LOG.debug("Creating new property collector.")
            property_collector = self.session._call_method(
                vim_util, "create_property_collector")
            self._register_vm_for_updates(vm_mor, property_collector)
            version = ""
            pg_key, port_key, swuuid = (None, None, None)
            while self.state == constants.DRIVER_RUNNING:
                LOG.debug("Waiting for VM %(vm)s to connect to "
                          "port group %(pg)s.",
                          {'vm': vm_mor.value, 'pg': pgmor.value})
                try:
                    update_set = self.session._call_method(
                        vim_util, "wait_for_updates_ex", version,
                        collector=property_collector)
                except error_util.SocketTimeoutException:
                    LOG.exception(_("Socket Timeout Exception."))
                    # Ignore timeout.
                    continue
                if update_set:
                    version = update_set.version
                    filterSet = update_set.filterSet
                    if not filterSet:
                        continue
                    for propFilterUpdate in filterSet:
                        objectSet = propFilterUpdate.objectSet
                        if not objectSet:
                            continue
                        for objectUpdate in objectSet:
                            if objectUpdate.kind == "leave":
                                LOG.warn(_("VM %(vm)s got deleted while "
                                           "waiting for it to connect to "
                                           "port group %(pg)s."),
                                         {'vm': vm_mor.value,
                                          'pg': pgmor.value})
                                return (pg_key, port_key, swuuid)
                            changes = common_util.convert_objectupdate_to_dict(
                                objectUpdate)
                            devices = changes.get('config.hardware.device')
                            nicdvs = network_util.get_vnics_from_devices(
                                devices)
                            if not nicdvs:
                                continue
                            for device in nicdvs:
                                if (hasattr(device, "backing") and
                                        hasattr(device.backing, "port") and
                                        device.backing.port):
                                    port = device.backing.port
                                    if (hasattr(port, "portgroupKey")):
                                        pg_key = port.portgroupKey
                                        if (pg_key == pgmor.value and
                                                hasattr(port, "portKey")):
                                            port_key = port.portKey
                                            swuuid = port.switchUuid
                                            LOG.info(_("VM %(vm)s connected "
                                                       "to port group: "
                                                       "%(pg)s."),
                                                     {'vm': vm_mor.value,
                                                      'pg': pgmor.value})
                                            return (pg_key, port_key, swuuid)
        except Exception as e:
            LOG.exception(_("Exception while waiting for VM %(vm)s "
                            "to connect to port group %(pg)s: %(err)s."),
                          {'vm': vm_mor.value, 'pg': pgmor.value, 'err': e})
            raise e
        finally:
            LOG.debug("Destroying the property collector created.")
            self.session._call_method(vim_util,
                                      "destroy_property_collector",
                                      property_collector)

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def create_network(self, network, virtual_switch):
        LOG.info(_("Creating portgroup %(nm)s with vlan id %(vid)s "
                   "on virtual switch %(sw)s."),
                 {'nm': network.name, 'vid': network.config.vlan.vlanIds[0],
                 'sw': virtual_switch.name})
        network_util.create_port_group(self.session,
                                       dvs_name=virtual_switch.name,
                                       pg_name=network.name,
                                       vlan_id=network.config.vlan.vlanIds[0])

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def get_pg_vlanid(self, dvs_name, pg_name):
        LOG.info(_("Fetching details of %(pg)s on %(dvs)s."),
                 {'pg': pg_name, 'dvs': dvs_name})
        pg_vlan_id = network_util.get_portgroup_details(self.session,
                                                        dvs_name, pg_name)
        return pg_vlan_id

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def get_vm_ref_uuid(self, vm_uuid):
        LOG.info(_("Fetching reference for %s."), vm_uuid)
        vm_ref = resource_util.get_vm_reference(self.session, vm_uuid)
        return vm_ref

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def wait_for_portgroup(self, vm_ref, pg_name):
        LOG.info(_("Wait for port group %s."), pg_name)
        pg_exists = network_util.wait_on_dvs_portgroup(self.session,
                                                       vm_ref, pg_name)
        return pg_exists

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def delete_network(self, network, virtual_switch=None):
        if not virtual_switch:
            dvs_list = cache.VCCache.get_cluster_switch_mapping().values()
            for dvs in dvs_list:
                self.delete_portgroup(switch=dvs,
                                      pg=network.name)

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def update_port(self, network=None, port=None, virtual_nic=None):
        device_id = port.vm_id
        mac_address = port.mac_address
        vm_mor = resource_util.get_vm_mor_for_uuid(self.session, device_id)
        if not vm_mor:
            LOG.warn(_("VM %(vm)s with mac address %(mac)s for port %(uuid)s "
                       "not found on this node."),
                     {'vm': device_id, 'mac': mac_address, 'uuid': port.uuid})
            return False
        if port.port_status == constants.PORT_STATUS_UP:
            enabled = True
        elif port.port_status == constants.PORT_STATUS_DOWN:
            enabled = False
        else:
            msg = (_("Invalid port status %(port)s in update for port %(id)s"),
                   {'port': port.port_status, 'id': port.uuid})
            raise error.OVSvAppNeutronAgentError(msg)
        action = "Enabling" if enabled else "Disabling"
        LOG.debug("%(action)s port used by VM %(id)s for VNIC with "
                  "mac address %(mac)s.",
                  {'action': action, 'id': device_id, 'mac': mac_address})
        status = network_util.enable_disable_port_of_vm(self.session,
                                                        vm_mor,
                                                        mac_address,
                                                        enabled)
        return status

    @utils.require_state(state=[constants.DRIVER_READY,
                                constants.DRIVER_RUNNING])
    def post_create_port(self, port):
        if port.port_status == constants.PORT_STATUS_UP:
            device_id = port.vm_id
            _clu_mor, _clu_path, vds_name = self._find_cluster_switch_for_vm(
                device_id)
            pg_mor = network_util.get_portgroup_mor_by_name(self.session,
                                                            vds_name,
                                                            port.network_uuid)
            if pg_mor is None:
                msg = (_("Port group %(net_id)s not created on "
                         "virtual switch %(vds)s."),
                       {'net_id': port.network_uuid,
                        'vds': vds_name})
                raise error_util.RunTimeError(msg)
            vm_mor = resource_util.get_vm_mor_for_uuid(self.session,
                                                       device_id)
            if vm_mor is None:
                msg = (_("Virtual machine %(id)s with "
                         "port %(port)s not created."),
                       {'id': device_id,
                        'port': port.uuid})
                raise error_util.RunTimeError(msg)
            (pg_key, port_key, swuuid) = self._wait_for_port_update_on_vm(
                vm_mor, pg_mor)
            if pg_key and port_key and swuuid:
                # enable the port on virtual switch.
                network_util.enable_disable_port(self.session, swuuid,
                                                 pg_key, port_key, True)


# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_vsphere.tests import base
from networking_vsphere.tests.unit.utils import fake_vmware_api
from networking_vsphere.utils import vim_util


class VimUtilsTestCase(base.TestCase):

    def setUp(self):
        super(VimUtilsTestCase, self).setUp()

    def test_build_recursive_traversal_spec(self):
        client_factory = fake_vmware_api.FakeFactory()
        trav_specs = vim_util.build_recursive_traversal_spec(client_factory)
        spec_names = ["rpToRp", "rpToVm", "crToRp", "crToH",
                      "dcToHf", "dcToVmf", "dcToDs", "hToVm",
                      "dsToVm", "visitFolders"]
        for spec in trav_specs:
            self.assertIn(spec.name, spec_names)

    def test_dynamic_property(self):
        vim = fake_vmware_api.FakeVimUtil()
        mob = {"_type": "vm"}
        vim_util.get_dynamic_property(vim, mob, "VM", "runtime.host")
        self.assertTrue(vim.RetrievePropertiesExCalled)
        self.assertTrue(vim.ContinueRetrievePropertiesExCalled)

    def test_dynamic_property_none_object(self):
        vim = fake_vmware_api.FakeVimUtil()
        vim_util.get_dynamic_property(vim, None, "VM", "runtime.host")
        self.assertFalse(vim.RetrievePropertiesExCalled)

    def test_get_property_filter_specs(self):
        vim = fake_vmware_api.FakeVimUtil()
        property_dict = {"virtualmachine":
                         ["name", "config"]}
        property_filter_spec = vim_util.get_property_filter_specs(
            vim, property_dict, None)
        objSpec = property_filter_spec.objectSet[0]
        self.assertEqual(objSpec.obj, vim.service_content.rootFolder)

    def test_get_properties_for_a_collection_of_objects(self):
        vim = fake_vmware_api.FakeVimUtil()
        obj_list = [fake_vmware_api.FakeDynamicPropertyObject]
        properties = ["name", "config"]
        objs = vim_util.get_properties_for_a_collection_of_objects(
            vim, "FakeDynamicPropertyObject", obj_list, properties)
        self.assertEqual(len(objs), 1)
        self.assertTrue(vim.RetrievePropertiesExCalled)
        for obj in objs:
            for prop in obj.propSet:
                self.assertIn(prop.name, properties)

    def test_get_properties_for_a_collection_empty(self):
        vim = fake_vmware_api.FakeVimUtil()
        obj_list = []
        properties = ["name", "config"]
        objs = vim_util.get_properties_for_a_collection_of_objects(
            vim, "FakeDynamicPropertyObject", obj_list, properties)
        self.assertEqual(len(objs), 0)
        self.assertFalse(vim.RetrievePropertiesExCalled)

    def test_get_objects(self):
        vim = fake_vmware_api.FakeVimUtil()
        vim_util.get_objects(vim, "DataCenter")
        self.assertTrue(vim.RetrievePropertiesExCalled)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from eventlet import greenthread
from oslo_config import cfg
from oslo_log import log

from networking_vsphere.common import constants
from networking_vsphere.utils import common_util
from networking_vsphere.utils import error_util
from networking_vsphere.utils import resource_util
from networking_vsphere.utils import vim_util

LOG = log.getLogger(__name__)


def get_dvs_mor_by_uuid(session, uuid):
    """Returns DVS mor by UUID."""
    return session._call_method(vim_util,
                                "get_dvs_mor_by_uuid",
                                uuid)


def get_dvs_mor_by_name(session, dvs_name):
    """Returns DVS mor from its name."""
    dvs_mors = session._call_method(
        vim_util, "get_objects", "DistributedVirtualSwitch", ["name"])
    for dvs_mor in dvs_mors:
        propset_dict = common_util.convert_propset_to_dict(dvs_mor.propSet)
        if propset_dict['name'] == dvs_name:
            return dvs_mor.obj
    return None


def get_all_portgroup_mors_for_switch(session, dvs_name):
    """Returns a list of mors of all portgroups attached to specified DVS."""
    dvs_mor = get_dvs_mor_by_name(session, dvs_name)
    if dvs_mor:
        dvs_config = session._call_method(
            vim_util, "get_dynamic_property", dvs_mor,
            "DistributedVirtualSwitch", "portgroup")
        port_group_mors = dvs_config.ManagedObjectReference
        return port_group_mors
    return None


def get_unused_portgroup_names(session, dvs_name):
    """Returns a list of unused portgroups.

    Unused portgroups - Portgroups of a specific DVS not
    connected to any virtual Machine.
    """
    unused_port_group_names = []
    port_group_mors = get_all_portgroup_mors_for_switch(session, dvs_name)
    if port_group_mors:
        port_groups = session._call_method(
            vim_util, "get_properties_for_a_collection_of_objects",
            "DistributedVirtualPortgroup", port_group_mors,
            ["summary.name", "tag", "vm"])
        for port_group in port_groups:
            propset_dict = common_util.convert_propset_to_dict(
                port_group.propSet)
            if not propset_dict['vm'] and not propset_dict['tag']:
                unused_port_group_names.append(propset_dict['summary.name'])
    return unused_port_group_names


def get_portgroup_mor_by_name(session, dvs_name, port_group_name):
    """Get Portgroup mor by Portgroup name."""
    port_group_mors = get_all_portgroup_mors_for_switch(session, dvs_name)
    if port_group_mors:
        port_groups = session._call_method(
            vim_util, "get_properties_for_a_collection_of_objects",
            "DistributedVirtualPortgroup", port_group_mors, ["summary.name"])
        for port_group in port_groups:
            if port_group.propSet[0].val == port_group_name:
                return port_group.obj
    return None


def _get_add_vswitch_port_group_spec(client_factory,
                                     port_group_name, vlan_id):
    """Builds DVS port group configuration spec."""
    vswitch_port_group_spec = client_factory.create(
        'ns0:DVPortgroupConfigSpec')
    vswitch_port_group_spec.name = port_group_name

    portSettingSpec = client_factory.create('ns0:VMwareDVSPortSetting')
    vlanSpec = client_factory.create(
        'ns0:VmwareDistributedVirtualSwitchVlanIdSpec')
    vlanSpec.inherited = False
    vlanSpec.vlanId = vlan_id
    portSettingSpec.vlan = vlanSpec

    vswitch_port_group_spec.autoExpand = True
    vswitch_port_group_spec.type = "earlyBinding"
    vswitch_port_group_spec.defaultPortConfig = portSettingSpec
    return vswitch_port_group_spec


def get_portgroup_details(session, dvs_name, pg_name):
    """Get VLAN id associated with a portgroup on a DVS."""
    port_group_mor = get_portgroup_mor_by_name(session, dvs_name, pg_name)
    vlan_id = constants.DEAD_VLAN
    if port_group_mor:
        port_group_config = session._call_method(
            vim_util, "get_dynamic_property", port_group_mor,
            "DistributedVirtualPortgroup", "config")
        vlan_id = port_group_config.defaultPortConfig.vlan.vlanId
    return vlan_id


def wait_until_dvs_portgroup_available(session, vm_ref, pg_name, wait_time):
    """Wait until a portgroup is available on a DVS."""
    time_elapsed = 0
    while time_elapsed < wait_time:
        host = session._call_method(vim_util, "get_dynamic_property",
                                    vm_ref, "VirtualMachine", "runtime.host")
        vm_networks_prop = session._call_method(vim_util,
                                                "get_dynamic_property", host,
                                                "HostSystem", "network")
        if vm_networks_prop:
            vm_networks = vm_networks_prop.ManagedObjectReference
            for network in vm_networks:
                if network._type == 'DistributedVirtualPortgroup':
                    props = session._call_method(vim_util,
                                                 "get_dynamic_property",
                                                 network,
                                                 network._type,
                                                 "config")
                    if props.name in pg_name:
                        LOG.debug("DistributedVirtualPortgroup created %s "
                                  % pg_name)
                        return True
        LOG.debug("Portgroup %s not created yet. Retrying again "
                  "after 5 seconds" % pg_name)
        greenthread.sleep(5)
        time_elapsed += 5
    if time_elapsed >= wait_time:
        LOG.debug("Portgroup %(pg)s not created within %(secs)s secs",
                  {'pg': pg_name, 'secs': wait_time})
    return False


def create_port_group(session, dvs_name, pg_name, vlan_id):
    """Creates a Portgroup on DVS with a vlan id."""
    port_group_mor = get_portgroup_mor_by_name(session, dvs_name, pg_name)
    if port_group_mor:
        port_group_config = session._call_method(
            vim_util, "get_dynamic_property", port_group_mor,
            "DistributedVirtualPortgroup", "config")
        if vlan_id == port_group_config.defaultPortConfig.vlan.vlanId:
            LOG.debug("Portgroup %(pg)s with vlan id %(vid)s already exists",
                      {'pg': pg_name, 'vid': vlan_id})
            return
        else:
            LOG.info(_("Portgroup %(pg)s already exists "
                     "but with vlan id %(vid)s"),
                     {'pg': pg_name,
                      'vid': port_group_config.defaultPortConfig.vlan.vlanId})
            raise error_util.RunTimeError("Inconsistent vlan id for portgroup"
                                          " %s", pg_name)
    else:
        client_factory = session._get_vim().client.factory
        add_prt_grp_spec = _get_add_vswitch_port_group_spec(
            client_factory, pg_name, vlan_id)
        blocked = client_factory.create('ns0:BoolPolicy')
        blocked.value = False
        blocked.inherited = False
        add_prt_grp_spec.defaultPortConfig.blocked = blocked
        dvs_mor = get_dvs_mor_by_name(session, dvs_name)

        try:
            task_ref = session._call_method(
                session._get_vim(), "AddDVPortgroup_Task", dvs_mor,
                spec=add_prt_grp_spec)
            session.wait_for_task(task_ref)
            LOG.info(_("Successfully created portgroup "
                     "%(pg)s with vlan id %(vid)s"),
                     {'pg': pg_name, 'vid': vlan_id})
        except Exception as e:
            LOG.exception(_("Failed to create portgroup %(pg)s with "
                          "vlan id %(vid)s on vCenter. Cause : %(err)s"),
                          {'pg': pg_name, 'vid': vlan_id, 'err': e})
            raise error_util.RunTimeError("Failed to create portgroup %s "
                                          "with vlan id %s on vCenter.Cause"
                                          " : %s" % (pg_name, vlan_id, e))


def delete_port_group(session, dvs_name, pg_name):
    """Deletes a port group from DVS."""
    port_group_mor = get_portgroup_mor_by_name(session, dvs_name, pg_name)
    if port_group_mor:
        try:
            destroy_task = session._call_method(session._get_vim(),
                                                "Destroy_Task", port_group_mor)
            session.wait_for_task(destroy_task)
            LOG.info(_("Successfully deleted portgroup %(pg)s from "
                     "dvs %(dvs)s"),
                     {'pg': pg_name, 'dvs': dvs_name})
        except Exception as e:
            LOG.exception(_("Failed to delete portgroup %(pg)s from "
                          "dvs %(dvs)s .Cause : %(err)s"),
                          {'pg': pg_name, 'dvs': dvs_name, 'err': e})
            raise error_util.RunTimeError("Failed to delete portgroup %s "
                                          "on dvs %s on vCenter.Cause"
                                          " : %s" % (pg_name, dvs_name, e))
    else:
        LOG.info(_("portgroup %(pg)s not present on dvs %(dvs)s"),
                 {'pg': pg_name, 'dvs': dvs_name})


def enable_disable_port_of_vm(session, vm_mor, mac_address, enabled):
    """Enable or disable a port of a VM having specific mac_address."""
    props = session._call_method(vim_util,
                                 "get_dynamic_properties",
                                 vm_mor,
                                 ["config.hardware.device"])
    devices = props["config.hardware.device"]
    LOG.debug("Found %(nod)s devices on VM %(vm)s",
              {'nod': len(devices.VirtualDevice), 'vm': vm_mor.value})
    vnics = get_vnics_from_devices(devices)
    for device in vnics:
        if (hasattr(device, "macAddress") and
                device.macAddress == mac_address):
            port = device.backing.port
            pgkey = port.portgroupKey
            portkey = port.portKey
            swuuid = port.switchUuid
            enable_disable_port(session, swuuid, pgkey, portkey, enabled)
            return True
    return False


def get_vnics_from_devices(devices):
    """Obtain vnic information."""
    vnics = None
    if (devices and hasattr(devices, "VirtualDevice")):
        vnics = []
        devices = devices.VirtualDevice
        for device in devices:
            if (device.__class__.__name__ in
                ("VirtualEthernetCard",
                 "VirtualE1000", "VirtualE1000e",
                 "VirtualPCNet32", "VirtualVmxnet",
                 "VirtualVmxnet2", "VirtualVmxnet3")):
                vnics.append(device)
    return vnics


def enable_disable_port(session, swuuid, pgkey, portkey, enabled):
    """Enable or Disable VM port."""
    action = "Enabling" if enabled else "Disabling"
    LOG.debug("%(action)s port %(port)s on %(pg)s",
              {'action': action, 'port': portkey, 'pg': pgkey})
    vds_mor = get_dvs_mor_by_uuid(session, swuuid)
    client_factory = session._get_vim().client.factory
    spec = client_factory.create('ns0:DVPortConfigSpec')
    spec.key = portkey
    spec.operation = "edit"
    setting = client_factory.create('ns0:DVPortSetting')
    blocked = client_factory.create('ns0:BoolPolicy')
    blocked.value = not enabled
    blocked.inherited = False
    setting.blocked = blocked
    spec.setting = setting
    reconfig_task = session._call_method(session._get_vim(),
                                         "ReconfigureDVPort_Task",
                                         vds_mor,
                                         port=[spec])
    session.wait_for_task(reconfig_task)
    action = "enabled" if enabled else "disabled"
    LOG.debug("Successfully %(action)s port %(port)s on port "
              "group %(pg)s on dvs %(dvs)s",
              {'action': action, 'port': portkey, 'pg': pgkey, 'dvs': swuuid})


def is_valid_dvswitch(session, cluster_mor, dvs_name):
    """Validate a DVS.

    Check if DVS exists for a cluster specified in conf.
    Also validate if DVS is attached to all hosts in a cluster.
    """
    dvs_mor = get_dvs_mor_by_name(session, dvs_name)
    if dvs_mor:
        dvs_config = session._call_method(
            vim_util, "get_dynamic_property", dvs_mor,
            "DistributedVirtualSwitch", "config.host")
        # Get all the host attached to given VDS
        dvs_host_members = dvs_config[0]
        dvs_attached_host_ids = []
        for dvs_host_member in dvs_host_members:
            dvs_attached_host_ids.append(dvs_host_member.config.host.value)

        # Get all the hosts present in the cluster
        hosts_in_cluster = resource_util.get_host_mors_for_cluster(
            session, cluster_mor)

        # Check if the host on which OVSvApp VM is hosted is a part of DVSwitch
        if hosts_in_cluster:
            for host in hosts_in_cluster:
                hostname = resource_util.get_hostname_for_host_mor(
                    session, host)
                if hostname == cfg.CONF.VMWARE.esx_hostname:
                    if host.value not in dvs_attached_host_ids:
                        LOG.error(_("DVS not present on host %s") % host.value)
                        return False
            return hosts_in_cluster
    else:
        LOG.error(_("DVS not present %s") % dvs_name)
        return False



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import contextlib

import mock
from oslo_config import cfg

from neutron.common import topics
from neutron.extensions import portbindings
from neutron.plugins.ml2 import driver_api as api
from neutron.tests.unit.plugins.ml2 import _test_mech_agent as base
from neutron.tests.unit.plugins.ml2 import test_rpc

from networking_vsphere.agent import ovsvapp_agent
from networking_vsphere.common import constants as ovsvapp_const
from networking_vsphere.ml2 import ovsvapp_rpc

cfg.CONF.import_group('ml2', 'neutron.plugins.ml2.config')

FAKE_CLUSTER_ID = 'fake_cluster_id'
FAKE_VCENTER = 'fake_vcenter'
FAKE_HOST = 'fake_host'
FAKE_AGENT_ID = 'fake_agent_id'
FAKE_PORT_ID = 'fake_port_id'
FAKE_NETWORK_ID = "fake_network_id"
FAKE_SUBNET_ID = "fake_subnet_id"
FAKE_DEVICE_OWNER = "fake_device_owner"
FAKE_DEVICE_ID = "fake_device_id"
FAKE_MAC_ADDRESS = "fake_mac_address"
FAKE_IP_ADDRESS = "fake_ip_address"


class TestFakePortContext(base.FakePortContext):
    VIF_TYPE = portbindings.VIF_TYPE_OTHER
    VIF_DETAILS = {portbindings.CAP_PORT_FILTER: True}
    AGENT_TYPE = ovsvapp_const.AGENT_TYPE_OVSVAPP

    GOOD_MAPPINGS = {'fake_physical_network': 'fake_bridge'}
    GOOD_TUNNEL_TYPES = ['vxlan']
    GOOD_CONFIGS = {'bridge_mappings': GOOD_MAPPINGS,
                    'tunnel_types': GOOD_TUNNEL_TYPES}
    AGENTS = [{'alive': True,
               'configurations': GOOD_CONFIGS}]

    def __init__(self, port, segments=None):
        agent_type = self.AGENT_TYPE
        agents = self.AGENTS
        vnic_type = portbindings.VNIC_NORMAL
        super(TestFakePortContext, self).__init__(agent_type,
                                                  agents,
                                                  segments,
                                                  vnic_type)
        if segments:
            self._bound_segment_id = segments[0].get(api.ID)
            self._bound_vif_type = 'other'
            self._bound_segment = segments[0]

        self._bound_segment = None
        self._port = port
        self._segments = segments

    @property
    def current(self):
        return self._port

    @property
    def status(self):
        return 'ACTIVE'

    @property
    def top_bound_segment(self):
        if self._segments:
            return self._expand_segment(self._bound_segment)


class OVSvAppServerRpcCallbackTest(test_rpc.RpcCallbacksTestCase):

    def setUp(self):
        super(OVSvAppServerRpcCallbackTest, self).setUp()
        self.ovsvapp_callbacks = ovsvapp_rpc.OVSvAppServerRpcCallback(
            mock.Mock())
        self.plugin = self.manager.get_plugin()

    def test_get_ports_for_device(self):
        kwargs = {'agent_id': FAKE_AGENT_ID,
                  'host': FAKE_HOST,
                  'device': {'id': 1,
                             'cluster_id': FAKE_CLUSTER_ID,
                             'vcenter': FAKE_VCENTER}}

        port = collections.defaultdict(lambda: 'fake')
        network = collections.defaultdict(lambda: 'fake')
        port['id'] = FAKE_PORT_ID
        port['security_groups'] = ['fake-sg1', 'fake-sg2']
        with contextlib.nested(
            mock.patch.object(self.plugin,
                              'get_ports',
                              return_value=[port]),
            mock.patch.object(self.plugin,
                              'get_network',
                              return_value=network),
            mock.patch.object(self.ovsvapp_callbacks.notifier,
                              'device_create'),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug'),
            mock.patch.object(self.plugin, 'get_ports_from_devices'),
            mock.patch.object(self.plugin, 'security_group_rules_for_ports'),
            mock.patch.object(self.ovsvapp_callbacks, 'update_port_binding',
                              return_value=[port])
        ) as (get_ports, get_network, device_create,
              log_debug, get_ports_from_devices, sg_rules_for_ports,
              update_port_binding):
            self.assertTrue(self.ovsvapp_callbacks.get_ports_for_device(
                            'fake_context', **kwargs))
            self.assertTrue(device_create.called)
            self.assertEqual(2, log_debug.call_count)
            self.assertTrue(get_ports_from_devices.called)
            self.assertTrue(sg_rules_for_ports.called)
            self.assertTrue(update_port_binding.called)

    def test_get_ports_for_device_no_security_groups(self):
        kwargs = {'agent_id': FAKE_AGENT_ID,
                  'host': FAKE_HOST,
                  'device': {'id': 1,
                             'cluster_id': FAKE_CLUSTER_ID,
                             'vcenter': FAKE_VCENTER}}

        port = collections.defaultdict(lambda: 'fake')
        network = collections.defaultdict(lambda: 'fake')
        port['id'] = 'fake-id'
        with contextlib.nested(
            mock.patch.object(self.plugin,
                              'get_ports',
                              return_value=[port]),
            mock.patch.object(self.plugin,
                              'get_network',
                              return_value=network),
            mock.patch.object(self.ovsvapp_callbacks.notifier,
                              'device_create'),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug'),
            mock.patch.object(self.plugin, 'get_ports_from_devices'),
            mock.patch.object(self.ovsvapp_callbacks, 'update_port_binding',
                              return_value=[port])
        ) as (get_ports, get_network, device_create,
              log_debug, get_ports_from_devices, update_port_binding):
            self.assertTrue(self.ovsvapp_callbacks.get_ports_for_device(
                            'fake_context', **kwargs))
            self.assertTrue(device_create.called)
            self.assertEqual(2, log_debug.call_count)
            self.assertFalse(get_ports_from_devices.called)
            self.assertTrue(update_port_binding.called)

    def test_get_ports_for_device_without_port(self):
        self.plugin.get_ports.return_value = None
        with contextlib.nested(
            mock.patch.object(ovsvapp_rpc.LOG, 'debug'),
            mock.patch.object(ovsvapp_rpc.LOG, 'exception')
        ) as (log_debug, log_exception):
            self.assertFalse(self.ovsvapp_callbacks.get_ports_for_device(
                             'fake_context', agent_id=FAKE_AGENT_ID,
                             host=FAKE_HOST,
                             device={'id': 1,
                                     'cluster_id': FAKE_CLUSTER_ID,
                                     'vcenter': FAKE_VCENTER}))
            self.assertEqual(2, log_debug.call_count)
            self.assertTrue(log_exception.called)

    def test_get_ports_for_device_without_device_id(self):
        kwargs = {'agent_id': FAKE_AGENT_ID,
                  'host': FAKE_HOST,
                  'device': {'id': None,
                             'cluster_id': FAKE_CLUSTER_ID,
                             'vcenter': FAKE_VCENTER}}
        with mock.patch.object(ovsvapp_rpc.LOG, 'debug') as log_debug:
            self.assertFalse(self.ovsvapp_callbacks.get_ports_for_device(
                             'fake_context', **kwargs))
            self.assertTrue(log_debug.called)

    def test_update_port_binding(self):
        kwargs = {'port_id': FAKE_PORT_ID, 'host': FAKE_HOST,
                  'agent_id': FAKE_AGENT_ID}
        port = {portbindings.HOST_ID: FAKE_HOST}
        self.plugin.update_port.return_value = port
        with mock.patch.object(ovsvapp_rpc.LOG, 'debug') as log_debug:
            updated_port = self.ovsvapp_callbacks.update_port_binding(
                'fake_context', **kwargs)
            self.assertEqual(port[portbindings.HOST_ID],
                             updated_port[portbindings.HOST_ID])
            self.assertTrue(log_debug.called)

    def test_update_ports_binding(self):
        kwargs = {'ports': set([FAKE_PORT_ID]), 'host': FAKE_HOST,
                  'agent_id': FAKE_AGENT_ID}
        port = {portbindings.HOST_ID: FAKE_HOST,
                'id': FAKE_PORT_ID}
        self.plugin.update_port.return_value = port
        with contextlib.nested(
            mock.patch.object(self.plugin, 'port_bound_to_host',
                              return_value=False),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug')
        ) as (port_bound_to_host, log_debug):
            status = self.ovsvapp_callbacks.update_ports_binding(
                'fake_context', **kwargs)
            self.assertIn("fake_port_id", status)
            self.assertTrue(log_debug.called)

    def test_update_ports_binding_no_host_change(self):
        kwargs = {'ports': set([FAKE_PORT_ID]), 'host': 'old_fake_host',
                  'agent_id': FAKE_AGENT_ID}
        port = {portbindings.HOST_ID: FAKE_HOST,
                'id': FAKE_PORT_ID}
        self.plugin.update_port.return_value = port
        with contextlib.nested(
            mock.patch.object(self.plugin, 'port_bound_to_host',
                              return_value=True),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug')
        ) as (get_host, log_debug):
            status = self.ovsvapp_callbacks.update_ports_binding(
                'fake_context', **kwargs)
            self.assertIn("fake_port_id", status)
            self.assertFalse(log_debug.called)

    def test_get_ports_details_list_all_ports_bound(self):
        kwargs = {'agent_id': FAKE_AGENT_ID,
                  'port_ids': [FAKE_PORT_ID],
                  'vcenter_id': FAKE_VCENTER,
                  'cluster_id': FAKE_CLUSTER_ID}
        fake_port_db = mock.Mock()
        fake_port_dict = {'id': FAKE_PORT_ID,
                          'fixed_ips': [{'subnet_id': FAKE_SUBNET_ID,
                                         'ip_address': FAKE_IP_ADDRESS}],
                          'device_id': FAKE_DEVICE_ID,
                          'device_owner': FAKE_DEVICE_OWNER,
                          'mac_address': FAKE_MAC_ADDRESS,
                          'admin_state_up': True,
                          'network_id': FAKE_NETWORK_ID}
        VLAN_SEGMENTS = [{api.ID: 'vlan_segment_id',
                          api.NETWORK_TYPE: 'vlan',
                          api.PHYSICAL_NETWORK: 'fake_physical_network',
                          api.SEGMENTATION_ID: 1234}]
        fake_port_context = TestFakePortContext(fake_port_dict, VLAN_SEGMENTS)
        fake_network = {'id': FAKE_NETWORK_ID}
        fake_context_obj = mock.Mock()
        with contextlib.nested(
            mock.patch('neutron.plugins.ml2.driver_context.PortContext',
                       return_value=fake_port_context),
            mock.patch.object(self.ovsvapp_callbacks,
                              '_get_port_db',
                              return_value=fake_port_db),
            mock.patch.object(self.plugin,
                              'get_network',
                              return_value=fake_network),
            mock.patch.object(self.plugin,
                              '_make_port_dict',
                              return_value=fake_port_dict),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug'),
        ) as (port_ctxt, get_port_db, get_network, make_port_dict, log_debug):
            actual = self.ovsvapp_callbacks.get_ports_details_list(
                fake_context_obj, **kwargs)
            self.assertEqual(1, get_port_db.call_count)
            self.assertEqual(1, get_network.call_count)
            self.assertEqual(2, log_debug.call_count)
            expected = [
                {'port_id': FAKE_PORT_ID,
                 'fixed_ips': [{'subnet_id': FAKE_SUBNET_ID,
                                'ip_address': FAKE_IP_ADDRESS}],
                 'device_id': FAKE_DEVICE_ID,
                 'device_owner': FAKE_DEVICE_OWNER,
                 'mac_address': FAKE_MAC_ADDRESS,
                 'lvid': 1234,
                 'admin_state_up': True,
                 'network_id': FAKE_NETWORK_ID,
                 'segmentation_id': VLAN_SEGMENTS[0][api.SEGMENTATION_ID],
                 'physical_network': VLAN_SEGMENTS[0][api.PHYSICAL_NETWORK],
                 'network_type': VLAN_SEGMENTS[0][api.NETWORK_TYPE]}]
            self.assertEqual(expected, actual)

    def test_get_ports_details_list_ports_not_bound(self):
        kwargs = {'agent_id': FAKE_AGENT_ID,
                  'port_ids': [FAKE_PORT_ID],
                  'vcenter_id': FAKE_VCENTER,
                  'cluster_id': FAKE_CLUSTER_ID}
        fake_port_db = mock.Mock()
        fake_port_dict = {'id': FAKE_PORT_ID,
                          'fixed_ips': [{'subnet_id': FAKE_SUBNET_ID,
                                         'ip_address': FAKE_IP_ADDRESS}],
                          'device_id': FAKE_DEVICE_ID,
                          'device_owner': FAKE_DEVICE_OWNER,
                          'mac_address': FAKE_MAC_ADDRESS,
                          'admin_state_up': True,
                          'network_id': FAKE_NETWORK_ID,
                          portbindings.VIF_TYPE: ''}
        fake_port_context = TestFakePortContext(fake_port_dict)
        fake_network = {'id': FAKE_NETWORK_ID}
        fake_context_obj = mock.Mock()
        with contextlib.nested(
            mock.patch('neutron.plugins.ml2.driver_context.PortContext',
                       return_value=fake_port_context),
            mock.patch.object(self.ovsvapp_callbacks,
                              '_get_port_db',
                              return_value=fake_port_db),
            mock.patch.object(self.plugin,
                              'get_network',
                              return_value=fake_network),
            mock.patch.object(self.plugin,
                              '_make_port_dict',
                              return_value=fake_port_dict),
            mock.patch.object(ovsvapp_rpc.LOG, 'debug'),
        ) as (port_ctxt, get_port_db, get_network, mk_port, log_debug):
            port_ctxt.return_value = fake_port_context
            actual = self.ovsvapp_callbacks.get_ports_details_list(
                fake_context_obj, **kwargs)
            self.assertEqual(1, get_port_db.call_count)
            self.assertEqual(1, get_network.call_count)
            self.assertEqual(1, log_debug.call_count)
            expected = []
            self.assertEqual(expected, actual)


class OVSvAppAgentNotifyAPITest(test_rpc.RpcApiTestCase):

    def test_device_create(self):
        rpcapi = ovsvapp_rpc.OVSvAppAgentNotifyAPI(topics.AGENT)
        self._test_rpc_api(rpcapi,
                           topics.get_topic_name(topics.AGENT,
                                                 ovsvapp_const.DEVICE,
                                                 topics.CREATE),
                           'device_create', rpc_method='cast',
                           fanout=True,
                           device='fake_device',
                           ports='fake_ports',
                           sg_rules='fake_sg_rules')

    def test_device_delete(self):
        rpcapi = ovsvapp_rpc.OVSvAppAgentNotifyAPI(topics.AGENT)
        self._test_rpc_api(rpcapi,
                           topics.get_topic_name(topics.AGENT,
                                                 ovsvapp_const.DEVICE,
                                                 topics.DELETE),
                           'device_delete', rpc_method='call',
                           network_info='fake_network_info',
                           host=FAKE_HOST)

    def test_get_ports_for_device(self):
        rpcapi = ovsvapp_agent.OVSvAppPluginApi(ovsvapp_const.OVSVAPP)
        self._test_rpc_api(rpcapi, None,
                           'get_ports_for_device', rpc_method='call',
                           device={'id': 'fake_id',
                                   'vcenter': FAKE_VCENTER,
                                   'cluster_id': FAKE_CLUSTER_ID},
                           agent_id=FAKE_AGENT_ID, host=FAKE_HOST)

    def test_update_port_binding(self):
        rpcapi = ovsvapp_agent.OVSvAppPluginApi(ovsvapp_const.OVSVAPP)
        self._test_rpc_api(rpcapi, None,
                           'update_port_binding', rpc_method='call',
                           port_id=FAKE_PORT_ID,
                           host=FAKE_HOST, agent_id=FAKE_AGENT_ID)

    def test_update_ports_binding(self):
        rpcapi = ovsvapp_agent.OVSvAppPluginApi(ovsvapp_const.OVSVAPP)
        self._test_rpc_api(rpcapi, None,
                           'update_ports_binding', rpc_method='call',
                           ports=['fake_ports'],
                           host=FAKE_HOST, agent_id=FAKE_AGENT_ID)

    def test_get_ports_details_list(self):
        rpcapi = ovsvapp_agent.OVSvAppPluginApi(ovsvapp_const.OVSVAPP)
        self._test_rpc_api(rpcapi, None,
                           'get_ports_details_list', rpc_method='call',
                           port_ids=['fake_port_ids'],
                           agent_id=FAKE_AGENT_ID, vcenter_id=FAKE_VCENTER,
                           cluster_id=FAKE_CLUSTER_ID)

    def test_update_local_vlan_assignment(self):
        rpcapi = ovsvapp_agent.OVSvAppPluginApi(ovsvapp_const.OVSVAPP)
        self._test_rpc_api(rpcapi, None,
                           'update_lvid_assignment', rpc_method='call',
                           net_info='fake_network_info')
